// CsParserTypes.h

#pragma once

#ifndef _CS_PARSER_TYPES_H_
#define _CS_PARSER_TYPES_H_



#include "CsToken.h"
#include "AstVisitor.h"
#include "SingleList.h"
#include "FastArray.h"
#include "FastDeque.h"

#include "CsEntity.h"



//------------------------------------------------------------------------------


//------------------------------------------------------------------------------


//------------------------------------------------------------------------------

#define DEFINE_AST_ALLOCATOR(Type)\
inline void * Type::operator new ( size_t size )\
{\
	return FixedSmallObjAllocator < ALIGN_TO_2_POW_N_OR_64 ( sizeof ( Type ) ), AstAllocatorId, OwnPageAllocator >::allocate();\
}\
\
inline void Type::operator delete ( void * address )\
{\
	FixedSmallObjAllocator < ALIGN_TO_2_POW_N_OR_64 ( sizeof ( Type ) ), AstAllocatorId, OwnPageAllocator >::deallocate ( address );\
}\

//------------------------------------------------------------------------------

#undef INVARIANT
	#define INVARIANT()

//------------------------------------------------------------------------------

#define DELETE_FROM_CONTAINER(cont_type,cont)\
{\
	for ( cont_type::iterator it = cont.begin(); it != cont.end(); ++it )\
		delete *it;\
}\

//------------------------------------------------------------------------------

class CsDisplay;

struct CsNode;
struct CsBlock;
struct CsExpressionList;
struct CsTypeRef;
struct CsExpression;
struct CsArgumentList;
struct CsArrayInitializer;
struct CsUsingDirective;
struct CsUserDefinedConversion;
class CsFile;

//------------------------------------------------------------------------------



//------------------------------------------------------------------------------
// CsTokenCollector
//------------------------------------------------------------------------------

class CsTokenCollector
{
public:
	inline void resize0()
	{
		tokens.resize0();
	}
	inline void push_back ( CsTokenIndex token )
	{
		if ( token )
		{


			tokens.push_back ( token );
		}
	}
	MediumFastDeque<CsTokenIndex> tokens;

};

//------------------------------------------------------------------------------
// CsEntityCollector
//------------------------------------------------------------------------------

class CsEntityCollector
{
public:
	inline void push_back ( CsEntity * entity )
	{
		if ( entity )
			entities.push_back ( entity );
	}
	MediumFastDeque<CsEntity*> entities;
};


//------------------------------------------------------------------------------
// cs_node
//------------------------------------------------------------------------------

enum cs_node
{
	n_null,
	
	n_using_alias_directive,
	n_using_namespace_directive,

// expressions

	n_simple_name,
	n_literal,
	n_embedded_type_member_access,
	n_this_access,
	n_base_member_access,
	n_base_indexer_access,
	n_typeof_expression,
	n_checked_expression,
	n_unchecked_expression,
	n_sizeof_expression,
	n_parenthesized_expression,
	n_new_object_expression,
	n_new_array_expression,
	n_invocation_expression,
	n_element_access,
	n_member_access,
	n_pointer_member_access,
	n_post_increment_decrement_expression,
	n_binary_expression,
	n_as_is_expression,
	n_conditional_expression,
	n_operator_unary_expression,
	n_pre_increment_decrement_expression,
	n_cast_unary_expression,
	n_assignment_expression,
	n_expression_list,
	n_makeref,
	n_reftype,
	n_refvalue,
	n_arglist,

// attributes

	n_attribute_named_argument,
	n_attribute,
	n_attribute_section,
	n_attributes,

// type reference

	n_type_ref,


// declarations

	n_type_base,
	n_argument_list,
	n_class,
	n_struct,
	n_interface,
	n_enum_member,
	n_enum,
	n_formal_parameter,
	n_formal_parameter_list,
	n_delegate,
	n_namespace,
	n_argument,

// statements

	n_label,
	n_labeled_statement,
	n_empty_statement,
	n_expression_statement,
	n_declaration_statement,
	n_local_constant_declarator,
	n_local_constant_declaration,
	n_if_statement,
	n_switch_label,
	n_switch_section,
	n_switch_statement,
	n_while_statement,
	n_do_while_statement,
	n_local_variable_declarator,
	n_local_variable_declaration,
	n_statement_expression_list,
	n_for_statement,
	n_foreach_statement,
	n_break_statement,
	n_continue_statement,
	n_goto_statement,
	n_return_statement,
	n_throw_statement,
	n_catch_clause,
	n_try_statement,
	n_checked_statement,
	n_unchecked_statement,
	n_lock_statement,
	n_using_statement,
	n_unsafe_statement,
	n_fixed_point_declarator,
	n_fixed_statement,
	n_stack_alloc_expression,
	n_block,

	// member declarations

	n_constructor_declaration,
	n_constant_declarator,
	n_constant_declaration,
	n_variable_declarator,
	n_variable_declaration,
	n_array_initializer,
	n_method_declaration,
	n_property_accessor,
	n_property_declaration,
	n_event_declaration,
	n_event_declarator,
	n_event_accessor,
	n_indexer_declaration,
	n_operator_declaration,
	n_conversion_operator_declaration,
	n_destructor_declaration,

	n_compilation_unit,

	n_end,
};

//------------------------------------------------------------------------------
// expression_classification
//------------------------------------------------------------------------------

enum expression_classification
{
	ec_unknown,
	ec_value,
	ec_variable,
	ec_namespace,
	ec_type,
	ec_method_group,
	ec_property_access,
	ec_indexer_access,
	ec_event_access,
	ec_nothing,
};

//------------------------------------------------------------------------------
// DeletedNodeInfo
//------------------------------------------------------------------------------


//------------------------------------------------------------------------------
// CsNode
//------------------------------------------------------------------------------

struct CsNode
{
	explicit CsNode ( cs_node type_ )
		: type ( type_ )
		, parent ( 0 )
	{
	}

	CsNode ( cs_node type_, CsNode * parent_ )
		: type ( type_ )
		, parent ( parent_ )
	{
		//XASSERT ( id != 586 );
		stop_by_id();
	}

	CsNode ( cs_node type_, LineInfo line_info_ )
		: type ( type_ )
		, line_info ( line_info_ )
		, parent ( 0 )
	{
		//XASSERT ( id != 586 );
	}

	CsNode ( cs_node type_, LineInfo line_info_, CsNode * parent_ )
		: type ( type_ )
		, line_info ( line_info_ )
		, parent ( parent_ )
	{
		//XASSERT ( id != 586 );
		stop_by_id();
	}

	void stop_by_id();

	virtual ~CsNode()
	{
	}

	virtual void build_entities ( CsDisplay & display ) = 0;
	virtual void type_lookup ( CsDisplay & display ) = 0;
	virtual void member_lookup ( CsDisplay & display ) = 0;


	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector ) =0;
	virtual void addTokens ( CsTokenCollector & tokenCollector ) =0;
	virtual void addEntityDefinitions ( CsEntityCollector & tokenCollector ) =0;
	virtual void addEntityReferences ( CsEntityCollector & tokenCollector ) =0;
	virtual void clearEntities() =0;


	virtual void visit ( CsAstVisitor * visitor ) =0;

	virtual void callback ( CsAstCallback * visitor ) =0;


	const char * getFilePath() const;



	cs_node type : 8;
	unsigned flags : 24;
	CsNode * parent;
	LineInfo line_info;

private:
	CsNode ( const CsNode & rhs );
	CsNode& operator= ( const CsNode & rhs );
};

//------------------------------------------------------------------------------
// CsScopeNode
//------------------------------------------------------------------------------

struct CsScopeNode : public CsNode
{
	CsScopeNode ( cs_node type )
		: CsNode ( type )
		, entity(0)
	{}

	CsScopeNode ( cs_node type_, CsNode * parent )
		: CsNode ( type, parent )
		, entity(0)
	{}

	CsScopeNode ( cs_node type, LineInfo line_info )
		: CsNode ( type, line_info )
		, entity(0)
	{}

	CsScopeNode ( cs_node type_, LineInfo line_info, CsNode * parent )
		: CsNode ( type, line_info, parent )
		, entity(0)
	{}

	virtual ~CsScopeNode()
	{}

	virtual void clearEntities();

	CsScopeEntity * entity;
};

//------------------------------------------------------------------------------
// CsIdentifier
//------------------------------------------------------------------------------

struct CsIdentifier
{
	CsIdentifier()
		: separator_token(0)
		, identifier_token(0)
	{}

	explicit CsIdentifier ( WPHashString identifier_  )
		: identifier ( identifier_ )
		, separator_token(0)
		, identifier_token(0)
	{}


	CsIdentifier (
			WPHashString identifier_
		, CsTokenIndex identifier_token_
		)
		: identifier ( identifier_ )
		, separator_token ( 0 )
		, identifier_token ( identifier_token_ )
	{}

	CsIdentifier (
			WPHashString identifier_
		, CsTokenIndex identifier_token_
		, CsTokenIndex separator_token_
		)
		: identifier ( identifier_ )
		, identifier_token ( identifier_token_ )
		, separator_token ( separator_token_ )
	{}

	inline void delete_value()
	{}

	PHashString identifier;
	CsTokenIndex separator_token;
	CsTokenIndex identifier_token;
};

//------------------------------------------------------------------------------
// CsCompoundIdentifier
//------------------------------------------------------------------------------

struct CsCompoundIdentifier
	: public value_single_list1<CsIdentifier>
{
	CsCompoundIdentifier()
	{}

	explicit CsCompoundIdentifier ( WPHashString phs )
		: value_single_list1<CsIdentifier> ( CsIdentifier ( phs ) )
	{}

	CsCompoundIdentifier (
	  WPHashString phs
	, CsTokenIndex identifier_token	)
	: value_single_list1<CsIdentifier>
		( CsIdentifier ( phs, identifier_token	) )
	{}

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );


	inline bool isCompound() const
	{
		return first_.next_ != 0;
	}

	inline WPHashString getFirstIdentifier() const
	{
		return first_.value_.identifier;
	}

	inline WPHashString getLastIdentifier() const
	{
		if ( last_ )
			return last_->value_.identifier;
		else
			return first_.value_.identifier;
	}
};


//------------------------------------------------------------------------------
// CsUsingDirective
//------------------------------------------------------------------------------

// abstract
struct CsUsingDirective : public CsNode
{
	explicit CsUsingDirective ( cs_node type )
		: CsNode ( type )
		, using_token(0)
		, semicolon_token(0)
	{}

	CsUsingDirective ( cs_node type, LineInfo line_info )
		: CsNode ( type, line_info )
		, using_token(0)
		, semicolon_token(0)
	{}

	CsTokenIndex using_token;
	CsTokenIndex semicolon_token;
};

//------------------------------------------------------------------------------
// CsUsingAliasDirective
//------------------------------------------------------------------------------

struct CsUsingAliasDirective : public CsUsingDirective
{
	CsUsingAliasDirective()
		: CsUsingDirective ( n_using_alias_directive )
		, namespace_or_type_entity(0)
		, identifier_token(0)
		, assign_token(0)
		, using_directive_used_in_type_lookup(0)
	{}

	explicit CsUsingAliasDirective ( LineInfo line_info )
		: CsUsingDirective ( n_using_alias_directive, line_info )
		, namespace_or_type_entity(0)
		, identifier_token(0)
		, assign_token(0)
		, using_directive_used_in_type_lookup(0)
	{}

	virtual ~CsUsingAliasDirective()
	{
	}

	virtual void build_entities ( CsDisplay & display ) {}
	virtual void type_lookup ( CsDisplay & display ) {}
	virtual void member_lookup ( CsDisplay & display ) {}

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void addEntityDefinitions ( CsEntityCollector & tokenCollector );
	virtual void addEntityReferences ( CsEntityCollector & tokenCollector );
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	PHashString identifier;
	CsCompoundIdentifier namespace_or_type_name;
	CsEntity * namespace_or_type_entity;

	CsTokenIndex identifier_token;
	CsTokenIndex assign_token;

	CsUsingAliasDirective * using_directive_used_in_type_lookup;

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );
};

DEFINE_AST_ALLOCATOR(CsUsingAliasDirective)

//------------------------------------------------------------------------------
// CsUsingNamespaceDirective
//------------------------------------------------------------------------------

struct CsUsingNamespaceDirective : public CsUsingDirective
{
	CsUsingNamespaceDirective()
		: CsUsingDirective ( n_using_namespace_directive )
		, namespace_or_type_entity(0)
		, using_directive_used_in_type_lookup(0)
	{}

	explicit CsUsingNamespaceDirective ( LineInfo line_info )
		: CsUsingDirective ( n_using_namespace_directive, line_info )
		, namespace_or_type_entity(0)
		, using_directive_used_in_type_lookup(0)
	{}

	virtual ~CsUsingNamespaceDirective()
	{
	}

	virtual void build_entities ( CsDisplay & display ) {}
	virtual void type_lookup ( CsDisplay & display ) {}
	virtual void member_lookup ( CsDisplay & display ) {}

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void addEntityDefinitions ( CsEntityCollector & tokenCollector );
	virtual void addEntityReferences ( CsEntityCollector & tokenCollector );
	virtual void clearEntities();

	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
	}


	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	CsCompoundIdentifier namespace_or_type_name;
	CsEntity * namespace_or_type_entity;

	CsUsingAliasDirective * using_directive_used_in_type_lookup;

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );
};

DEFINE_AST_ALLOCATOR(CsUsingNamespaceDirective)

//------------------------------------------------------------------------------
// CsScopeNodeWithUsingDirectives
//------------------------------------------------------------------------------

struct CsScopeNodeWithUsingDirectives : public CsScopeNode
{
	explicit CsScopeNodeWithUsingDirectives ( cs_node type )
		: CsScopeNode ( type )
		, aliases(0)
		, bUsingDirectivesResolved(false)
	{}

	CsScopeNodeWithUsingDirectives ( cs_node type_, CsNode * parent )
		: CsScopeNode ( type, parent )
		, bUsingDirectivesResolved(false)
	{}

	CsScopeNodeWithUsingDirectives ( cs_node type, LineInfo line_info )
		: CsScopeNode ( type, line_info )
		, aliases(0)
		, bUsingDirectivesResolved(false)
	{}

	CsScopeNodeWithUsingDirectives ( cs_node type_, LineInfo line_info, CsNode * parent )
		: CsScopeNode ( type, line_info, parent )
		, bUsingDirectivesResolved(false)
	{}

	virtual ~CsScopeNodeWithUsingDirectives()
	{
		DELETE_FROM_CONTAINER(FastArray<CsUsingDirective*>,using_directives);
		if ( aliases )
			delete aliases;
	}

	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor )
	{
		FastArray<CsUsingDirective*>::iterator it = using_directives.begin();
		FastArray<CsUsingDirective*>::iterator it_end = using_directives.end();
		for ( ; it != it_end; ++it )
		{
			if ( *it )
				(*it)->visit ( visitor );
		}
	}

	FastArray<CsUsingDirective*> using_directives;
	CsScopeTable * aliases; // transformation parser store pointers to using directives (CsUsingDirective*) not entities
	bool bUsingDirectivesResolved;
};

//------------------------------------------------------------------------------
// TypeSpecifierEnum
//------------------------------------------------------------------------------

enum TypeSpecifierEnum
{
	//tsUnknown,
	tsArrayPostfix,
	tsPointer,
};

//------------------------------------------------------------------------------
// CsTypeSpecifier
//------------------------------------------------------------------------------

struct CsTypeSpecifier
{
	explicit CsTypeSpecifier ( TypeSpecifierEnum specifier )
		: specifier_type(specifier)
		, rank(0)
		, lbracket_star_token(0)
		, rbracket_token(0)
		, comma_tokens(0)
	{}

	CsTypeSpecifier ( TypeSpecifierEnum specifier, unsigned r )
		: specifier_type(specifier)
		, rank(r)
		, lbracket_star_token(0)
		, rbracket_token(0)
		, comma_tokens(0)
	{}

	CsTypeSpecifier (
			TypeSpecifierEnum specifier
		, unsigned r
		, CsTokenIndex lbracket_token
		, CsTokenIndex rbracket_token
		, value_single_list1<CsTokenIndex> * comma_tokens_
		)
		: specifier_type(specifier)
		, rank(r)
		, lbracket_star_token(lbracket_token)
		, rbracket_token(rbracket_token)
		, comma_tokens(comma_tokens_)
	{}

	void delete_value()
	{
		delete comma_tokens;
	}

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );

	TypeSpecifierEnum specifier_type;
	unsigned rank;
	CsTokenIndex lbracket_star_token; // '[' or '*'
	CsTokenIndex rbracket_token; // ']'
	value_single_list1<CsTokenIndex> * comma_tokens;
};

//------------------------------------------------------------------------------
// CsTypeSpecifiers
//------------------------------------------------------------------------------

struct CsTypeSpecifiers
{
	explicit CsTypeSpecifiers ( CsTypeSpecifier specifier )
		: specifiers ( specifier )
	{}

	virtual ~CsTypeSpecifiers()
	{
		//specifiers.delete_nodes();
	}

	virtual void build_entities ( CsDisplay & display ) {}
	virtual void type_lookup ( CsDisplay & display ) {}
	virtual void member_lookup ( CsDisplay & display ) {}

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );

	value_single_list1_with_delete_value<CsTypeSpecifier> specifiers;

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );
};

DEFINE_AST_ALLOCATOR(CsTypeSpecifiers)

//------------------------------------------------------------------------------
// CsTypeRef
//------------------------------------------------------------------------------

struct CsTypeRef : public CsNode
{
	CsTypeRef()
    : CsNode ( n_type_ref )
		, embedded_type(tkEOF)
		, specifiers(0)
		, separator_token(0)
		, embedded_type_token(0)
		, using_directive_used_in_type_lookup(0)
	{
	}

	explicit CsTypeRef ( LineInfo line_info )
		: CsNode ( n_type_ref, line_info )
		, embedded_type(tkEOF)
		, specifiers(0)
		, separator_token(0)
		, embedded_type_token(0)
		, using_directive_used_in_type_lookup(0)
	{
	}

	virtual ~CsTypeRef()
	{
		delete specifiers;
		entity_typeref.delete_entities();
	}

	virtual void build_entities ( CsDisplay & display ) {}
	virtual void type_lookup ( CsDisplay & display ) { ASSERT(0); }
	virtual void member_lookup ( CsDisplay & display ) {}

	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void addEntityDefinitions ( CsEntityCollector & tokenCollector ) {}
	virtual void addEntityReferences ( CsEntityCollector & tokenCollector ) {}
	virtual void clearEntities();



	inline bool isEntityType() const
	{
		return ( embedded_type == tkEOF );
	}

	CsTokenType embedded_type;
	CsCompoundIdentifier compound_identifier;
	CsTypeSpecifiers * specifiers;
	CsEntityTypeRef entity_typeref;
	CsTokenIndex separator_token;
	CsTokenIndex embedded_type_token;
	CsUsingAliasDirective * using_directive_used_in_type_lookup;

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );
};

DEFINE_AST_ALLOCATOR(CsTypeRef)

//------------------------------------------------------------------------------
// CsExpression
//------------------------------------------------------------------------------

// abstract
struct CsExpression : public CsNode
{
	explicit CsExpression ( cs_node type )
		: CsNode ( type )
		, ec ( ec_unknown )
		, bDeleteEntityTyperef(false)
		, entity_typeref(0)
		, entity(0)
		, user_defined_conversion(0)
		, separator_token(0)
	{
	}

	CsExpression ( cs_node type, LineInfo line_info )
		: CsNode ( type, line_info )
		, ec ( ec_unknown )
		, bDeleteEntityTyperef(false)
		, entity_typeref(0)
		, entity(0)
		, user_defined_conversion(0)
		, separator_token(0)
	{
	}

	CsExpression  ( cs_node type, LineInfo line_info, CsNode * parent )
		: CsNode ( type, line_info, parent )
		, ec ( ec_unknown )
		, bDeleteEntityTyperef(false)
		, entity_typeref(0)
		, entity(0)
		, user_defined_conversion(0)
		, separator_token(0)
	{
	}

	virtual ~CsExpression()
	{
		if ( bDeleteEntityTyperef )
		{
			entity_typeref->delete_entities();
			delete entity_typeref;
		}
	}

	virtual void build_entities ( CsDisplay & display ) {}


	virtual void addEntityDefinitions ( CsEntityCollector & entityCollector );
	virtual void addEntityReferences ( CsEntityCollector & entityCollector );
	virtual void clearEntities();


	expression_classification ec : 8;
	bool bDeleteEntityTyperef;
	CsEntityTypeRef * entity_typeref; // type of expression
	CsEntity * entity; // contains info about how to expression type was evaluated
	//CsUserDefinedConversion * conversion;
	CsEntityMethod * user_defined_conversion;
	CsTokenIndex separator_token;
};

//------------------------------------------------------------------------------
// CsPrimaryExpression
//------------------------------------------------------------------------------

// abstract
struct CsPrimaryExpression : public CsExpression
{
	explicit CsPrimaryExpression ( cs_node type )
		: CsExpression ( type )
	{}

	CsPrimaryExpression ( cs_node type, LineInfo line_info )
		: CsExpression ( type, line_info )
	{}

	CsPrimaryExpression ( cs_node type, LineInfo line_info, CsNode * parent )
		: CsExpression ( type, line_info, parent )
	{}

	virtual ~CsPrimaryExpression()
	{}
};

//------------------------------------------------------------------------------
// CsSimpleName
//------------------------------------------------------------------------------

struct CsSimpleName : public CsPrimaryExpression
{
	CsSimpleName()
		: CsPrimaryExpression ( n_simple_name )
		, identifier_token(0)
	{}

	explicit CsSimpleName ( LineInfo line_info )
		: CsPrimaryExpression ( n_simple_name, line_info )
		, identifier_token(0)
	{}

	virtual ~CsSimpleName()
	{
	}


	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );

	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	PHashString identifier;
	CsTokenIndex identifier_token;
};

DEFINE_AST_ALLOCATOR(CsSimpleName)

//------------------------------------------------------------------------------
// CsLiteral
//------------------------------------------------------------------------------

struct CsLiteral : public CsPrimaryExpression
{
	CsLiteral()
		: CsPrimaryExpression ( n_literal )
		, type(tkEOF)
		, literal_token(0)
	{}

	explicit CsLiteral ( LineInfo line_info )
		: CsPrimaryExpression ( n_literal, line_info )
		, type(tkEOF)
		, literal_token(0)
	{}

	virtual ~CsLiteral()
	{}

	virtual void type_lookup ( CsDisplay & display ) {}
	virtual void member_lookup ( CsDisplay & display );

	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	CsTokenType type;
	PHashString string;
	CsLiteralType literal_type;
	unsigned __int64 value;
	CsTokenIndex literal_token;
};

DEFINE_AST_ALLOCATOR(CsLiteral)

//------------------------------------------------------------------------------
// CsEmbeddedTypeMemberAccess
//------------------------------------------------------------------------------

struct CsEmbeddedTypeMemberAccess : public CsPrimaryExpression
{
	CsEmbeddedTypeMemberAccess()
		: CsPrimaryExpression ( n_embedded_type_member_access )
		, embedded_type ( tkEOF )
		, embedded_type_token(0)
		, dot_token(0)
		, member_token(0)
	{}

	explicit CsEmbeddedTypeMemberAccess ( LineInfo line_info )
		: CsPrimaryExpression ( n_embedded_type_member_access, line_info )
		, embedded_type ( tkEOF )
		, embedded_type_token(0)
		, dot_token(0)
		, member_token(0)
	{}

	virtual ~CsEmbeddedTypeMemberAccess()
	{
		embedded_entity_typeref.delete_entities();
	}

	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );


	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void addEntityDefinitions ( CsEntityCollector & tokenCollector ) {}
	virtual void addEntityReferences ( CsEntityCollector & tokenCollector ) {}
	virtual void clearEntities();

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	CsTokenType embedded_type;
	PHashString member;
	CsEntityTypeRef embedded_entity_typeref;
	CsTokenIndex embedded_type_token;
	CsTokenIndex dot_token;
	CsTokenIndex member_token;
};

DEFINE_AST_ALLOCATOR(CsEmbeddedTypeMemberAccess)


//------------------------------------------------------------------------------
// CsThisAccess
//------------------------------------------------------------------------------

struct CsThisAccess : public CsPrimaryExpression
{
	CsThisAccess()
		: CsPrimaryExpression ( n_this_access )
		, this_token(0)
	{}

	explicit CsThisAccess ( LineInfo line_info )
		: CsPrimaryExpression ( n_this_access, line_info )
		, this_token(0)
	{}

	virtual ~CsThisAccess()
	{}

	virtual void build_entities ( CsDisplay & display ) {}
	virtual void type_lookup ( CsDisplay & display ) {}
	virtual void member_lookup ( CsDisplay & display );

	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	CsTokenIndex this_token;
};

DEFINE_AST_ALLOCATOR(CsThisAccess)

//------------------------------------------------------------------------------
// CsBaseMemberAccess
//------------------------------------------------------------------------------

struct CsBaseMemberAccess : public CsPrimaryExpression
{
	CsBaseMemberAccess()
		: CsPrimaryExpression ( n_base_member_access )
		, base_token(0)
		, dot_token(0)
		, member_token(0)
	{}

	CsBaseMemberAccess ( LineInfo line_info )
		: CsPrimaryExpression ( n_base_member_access, line_info )
		, base_token(0)
		, dot_token(0)
		, member_token(0)
	{}

	virtual ~CsBaseMemberAccess()
	{
	}

	virtual void build_entities ( CsDisplay & display ) {}
	virtual void type_lookup ( CsDisplay & display ) {}
	virtual void member_lookup ( CsDisplay & display );

	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	PHashString member;
	CsTokenIndex base_token;
	CsTokenIndex dot_token;
	CsTokenIndex member_token;
};

DEFINE_AST_ALLOCATOR(CsBaseMemberAccess)

//------------------------------------------------------------------------------
// CsBaseIndexerAccess
//------------------------------------------------------------------------------

struct CsBaseIndexerAccess : public CsPrimaryExpression
{
	CsBaseIndexerAccess()
		: CsPrimaryExpression ( n_base_indexer_access )
		, expression_list(0)
		, base_token(0)
		, lbracket_token(0)
		, rbracket_token(0)
	{}

	explicit CsBaseIndexerAccess ( LineInfo line_info )
		: CsPrimaryExpression ( n_base_indexer_access, line_info )
		, expression_list(0)
		, base_token(0)
		, lbracket_token(0)
		, rbracket_token(0)
	{}

	virtual ~CsBaseIndexerAccess();

	virtual void build_entities ( CsDisplay & display ) {}
	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor );

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	CsExpressionList * expression_list;
	CsTokenIndex base_token;
	CsTokenIndex lbracket_token;
	CsTokenIndex rbracket_token;
};

DEFINE_AST_ALLOCATOR(CsBaseIndexerAccess)

//------------------------------------------------------------------------------
// CsTypeofExpression
//------------------------------------------------------------------------------

struct CsTypeofExpression : public CsPrimaryExpression
{
	CsTypeofExpression()
		: CsPrimaryExpression ( n_typeof_expression )
		, type(0)
		, typeof_token(0)
		, lparen_token(0)
		, rparen_token(0)
	{}

	explicit CsTypeofExpression ( LineInfo line_info )
		: CsPrimaryExpression  ( n_typeof_expression, line_info )
		, type(0)
		, typeof_token(0)
		, lparen_token(0)
		, rparen_token(0)
	{}

	virtual ~CsTypeofExpression()
	{
		delete type;
	}

	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor );

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	CsTypeRef * type;
	CsTokenIndex typeof_token;
	CsTokenIndex lparen_token;
	CsTokenIndex rparen_token;
};

DEFINE_AST_ALLOCATOR(CsTypeofExpression)

//------------------------------------------------------------------------------
// CsCheckedExpression
//------------------------------------------------------------------------------

struct CsCheckedExpression : public CsPrimaryExpression
{
	CsCheckedExpression()
		: CsPrimaryExpression ( n_checked_expression )
		, expression(0)
		, checked_token(0)
		, lparen_token(0)
		, rparen_token(0)
	{}

	explicit CsCheckedExpression ( LineInfo line_info )
		: CsPrimaryExpression ( n_checked_expression, line_info )
		, expression(0)
		, checked_token(0)
		, lparen_token(0)
		, rparen_token(0)
	{}

	virtual ~CsCheckedExpression()
	{
		delete expression;
	}

	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
		if ( expression )
			expression->visit ( visitor );
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	CsExpression * expression;
	CsTokenIndex checked_token;
	CsTokenIndex lparen_token;
	CsTokenIndex rparen_token;
};

DEFINE_AST_ALLOCATOR(CsCheckedExpression)

//------------------------------------------------------------------------------
// CsUncheckedExpression
//------------------------------------------------------------------------------

struct CsUncheckedExpression : public CsPrimaryExpression
{
	CsUncheckedExpression()
		: CsPrimaryExpression ( n_unchecked_expression )
		, expression(0)
		, unchecked_token(0)
		, lparen_token(0)
		, rparen_token(0)
	{}

	explicit CsUncheckedExpression ( LineInfo line_info )
		: CsPrimaryExpression ( n_unchecked_expression, line_info )
		, expression(0)
		, unchecked_token(0)
		, lparen_token(0)
		, rparen_token(0)
	{}

	virtual ~CsUncheckedExpression()
	{
		delete expression;
	}

	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
		if ( expression )
			expression->visit ( visitor );
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	CsExpression * expression;
	CsTokenIndex unchecked_token;
	CsTokenIndex lparen_token;
	CsTokenIndex rparen_token;
};

DEFINE_AST_ALLOCATOR(CsUncheckedExpression)

//------------------------------------------------------------------------------
// CsSizeofExpression
//------------------------------------------------------------------------------

struct CsSizeofExpression : public CsPrimaryExpression
{
	CsSizeofExpression()
		: CsPrimaryExpression ( n_sizeof_expression )
		, type(0)
		, sizeof_token(0)
		, lparen_token(0)
		, rparen_token(0)
	{}

	explicit CsSizeofExpression ( LineInfo line_info )
		: CsPrimaryExpression  ( n_sizeof_expression, line_info )
		, type(0)
		, sizeof_token(0)
		, lparen_token(0)
		, rparen_token(0)
	{}

	virtual ~CsSizeofExpression()
	{
		delete type;
	}

	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor );

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	CsTypeRef * type;
	CsTokenIndex sizeof_token;
	CsTokenIndex lparen_token;
	CsTokenIndex rparen_token;
};

DEFINE_AST_ALLOCATOR(CsSizeofExpression)

//------------------------------------------------------------------------------
// CsParenthesizedExpression
//------------------------------------------------------------------------------

struct CsParenthesizedExpression : public CsPrimaryExpression
{
	CsParenthesizedExpression()
		: CsPrimaryExpression ( n_parenthesized_expression )
		, expression(0)
		, lparen_token(0)
		, rparen_token(0)
	{}

	explicit CsParenthesizedExpression ( LineInfo line_info )
		: CsPrimaryExpression ( n_parenthesized_expression, line_info )
		, expression(0)
		, lparen_token(0)
		, rparen_token(0)
	{}

	virtual ~CsParenthesizedExpression()
	{
		delete expression;
	}

	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
		if ( expression )
			expression->visit ( visitor );
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	CsExpression * expression;
	CsTokenIndex lparen_token;
	CsTokenIndex rparen_token;
};

DEFINE_AST_ALLOCATOR(CsParenthesizedExpression)

//------------------------------------------------------------------------------
// CsNewObjectExpression
//------------------------------------------------------------------------------

struct CsNewObjectExpression : public CsPrimaryExpression
{
	CsNewObjectExpression()
		: CsPrimaryExpression ( n_new_object_expression, line_info )
		, type(0)
		, argument_list(0)
		, new_token(0)
		, lparen_token(0)
		, rparen_token(0)
	{}

	explicit CsNewObjectExpression ( LineInfo line_info )
		: CsPrimaryExpression ( n_new_object_expression, line_info )
		, type(0)
		, argument_list(0)
		, new_token(0)
		, lparen_token(0)
		, rparen_token(0)
	{}

	virtual ~CsNewObjectExpression();

	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor );

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	CsTypeRef * type;
	CsArgumentList * argument_list;
	CsTokenIndex new_token;
	CsTokenIndex lparen_token;
	CsTokenIndex rparen_token;
};

DEFINE_AST_ALLOCATOR(CsNewObjectExpression)

//------------------------------------------------------------------------------
// CsNewArrayExpression
//------------------------------------------------------------------------------

struct CsNewArrayExpression : public CsPrimaryExpression
{
	CsNewArrayExpression()
		: CsPrimaryExpression ( n_new_array_expression )
		, type(0)
		, expressions(0)
		, specifiers(0)
		, initializer(0)
		, new_token(0)
		, lbracket_token(0)
		, rbracket_token(0)
	{}

	explicit CsNewArrayExpression ( LineInfo line_info )
		: CsPrimaryExpression ( n_new_array_expression, line_info )
		, type(0)
		, expressions(0)
		, specifiers(0)
		, initializer(0)
		, new_token(0)
		, lbracket_token(0)
		, rbracket_token(0)
	{}

	virtual ~CsNewArrayExpression();

	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor );

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	CsTypeRef * type; // non-array-type (typeref not include type specifiers)
	CsExpressionList * expressions;
	//std::vector<CsTypeSpecifier> specifiers;
	value_single_list1_with_delete_value<CsTypeSpecifier> * specifiers;
	CsArrayInitializer * initializer;
	CsTokenIndex new_token;
	CsTokenIndex lbracket_token;
	CsTokenIndex rbracket_token;
};

DEFINE_AST_ALLOCATOR(CsNewArrayExpression)

//------------------------------------------------------------------------------
// CsInvocationExpression
//------------------------------------------------------------------------------

struct CsInvocationExpression : public CsPrimaryExpression
{
	CsInvocationExpression()
		: CsPrimaryExpression ( n_invocation_expression )
		, expression(0)
		, argument_list(0)
		, lparen_token(0)
		, rparen_token(0)
	{}

	CsInvocationExpression ( CsExpression * expr, LineInfo line_info )
		: CsPrimaryExpression ( n_invocation_expression, line_info )
		, expression(expr)
		, argument_list(0)
		, lparen_token(0)
		, rparen_token(0)
	{}

	virtual ~CsInvocationExpression();

	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void addEntityDefinitions ( CsEntityCollector & tokenCollector ) {}
	virtual void addEntityReferences ( CsEntityCollector & tokenCollector ) {}
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor );

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	CsExpression * expression;
	CsArgumentList * argument_list;
	CsTokenIndex lparen_token;
	CsTokenIndex rparen_token;
};

DEFINE_AST_ALLOCATOR(CsInvocationExpression)

//------------------------------------------------------------------------------
// CsElementAccess
//------------------------------------------------------------------------------

struct CsElementAccess : public CsPrimaryExpression
{
	CsElementAccess()
		: CsPrimaryExpression ( n_element_access )
		, expression(0)
		, expressions(0)
		, lbracket_token(0)
		, rbracket_token(0)
	{}

	CsElementAccess ( CsExpression * expr, LineInfo line_info )
		: CsPrimaryExpression  ( n_element_access, line_info )
		, expression(expr)
		, expressions(0)
		, lbracket_token(0)
		, rbracket_token(0)
	{}

	virtual ~CsElementAccess();

	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void addEntityDefinitions ( CsEntityCollector & tokenCollector ) {}
	virtual void addEntityReferences ( CsEntityCollector & tokenCollector ) {}
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor );

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	CsExpression * expression;
	CsExpressionList * expressions;
	CsTokenIndex lbracket_token;
	CsTokenIndex rbracket_token;
};

DEFINE_AST_ALLOCATOR(CsElementAccess)

//------------------------------------------------------------------------------
// CsMemberAccess
//------------------------------------------------------------------------------

struct CsMemberAccess : public CsPrimaryExpression
{
	CsMemberAccess()
		: CsPrimaryExpression ( n_member_access )
		, expression(0)
		, dot_token(0)
		, member_token(0)
	{}

	CsMemberAccess ( CsExpression * expr, LineInfo line_info )
		: CsPrimaryExpression ( n_member_access, line_info )
		, expression(expr)
		, dot_token(0)
		, member_token(0)
	{}

	virtual ~CsMemberAccess()
	{
		delete expression;
	}

	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
		if ( expression )
			expression->visit ( visitor );
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	CsExpression * expression;
	PHashString member;
	CsTokenIndex dot_token;
	CsTokenIndex member_token;
};

DEFINE_AST_ALLOCATOR(CsMemberAccess)

//------------------------------------------------------------------------------
// CsPointerMemberAccess
//------------------------------------------------------------------------------

struct CsPointerMemberAccess : public CsPrimaryExpression
{
	CsPointerMemberAccess()
		: CsPrimaryExpression ( n_pointer_member_access )
		, expression(0)
		, arrow_token(0)
		, member_token(0)
	{}

	CsPointerMemberAccess ( CsExpression * expr, LineInfo line_info )
		: CsPrimaryExpression ( n_pointer_member_access, line_info )
		, expression(expr)
		, arrow_token(0)
		, member_token(0)
	{}

	virtual ~CsPointerMemberAccess()
	{
		delete expression;
	}

	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
		if ( expression )
			expression->visit ( visitor );
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	CsExpression * expression;
	PHashString member;
	CsTokenIndex arrow_token;
	CsTokenIndex member_token;
};

DEFINE_AST_ALLOCATOR(CsPointerMemberAccess)

//------------------------------------------------------------------------------
// CsPostIncrementDecrementExpression
//------------------------------------------------------------------------------

struct CsPostIncrementDecrementExpression : public CsPrimaryExpression
{
	CsPostIncrementDecrementExpression()
		: CsPrimaryExpression ( n_post_increment_decrement_expression )
		, expression ( 0 )
		, operator_token(0)
	{}

	CsPostIncrementDecrementExpression ( CsExpression * expr, LineInfo line_info )
		: CsPrimaryExpression ( n_post_increment_decrement_expression, line_info )
		, expression ( expr )
		, operator_token(0)
	{}

	virtual ~CsPostIncrementDecrementExpression()
	{
		delete expression;
	}

	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
		if ( expression )
			expression->visit ( visitor );
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	CsExpression * expression;
	CsTokenType oper;
	CsTokenIndex operator_token;
};

DEFINE_AST_ALLOCATOR(CsPostIncrementDecrementExpression)

//------------------------------------------------------------------------------
// CsUnaryExpression
//------------------------------------------------------------------------------

//abstract
struct CsUnaryExpression : public CsExpression
{
	explicit CsUnaryExpression ( cs_node type )
		: CsExpression ( type )
		, unary_expression(0)
	{}

	CsUnaryExpression ( cs_node type, LineInfo line_info )
		: CsExpression ( type, line_info )
		, unary_expression(0)
	{}

	CsUnaryExpression ( cs_node type, LineInfo line_info, CsNode * parent )
		: CsExpression ( type, line_info, parent )
		, unary_expression(0)
	{}

	CsUnaryExpression ( cs_node type, LineInfo line_info, CsNode * parent, CsExpression * unary_expression_ )
		: CsExpression ( type, line_info, parent )
		, unary_expression(unary_expression_)
	{}

	virtual ~CsUnaryExpression()
	{
		delete unary_expression;
	}

	virtual void type_lookup ( CsDisplay & display );


public:
	CsExpression * unary_expression; // unary or primary expression
};

//------------------------------------------------------------------------------
// CsBinaryExpression
//------------------------------------------------------------------------------

struct CsBinaryExpression : public CsExpression
{
	CsBinaryExpression()
		: CsExpression ( n_binary_expression )
		, lhs(0)
		, oper(tkEOF)
		, rhs(0)
		, operator_token(0)
	{}

	CsBinaryExpression ( CsExpression * lhs_, CsTokenType oper_, CsExpression * rhs_ )
		: CsExpression ( n_binary_expression, line_info )
		, lhs(lhs_)
		, oper(oper_)
		, rhs(rhs_)
		, operator_token(0)
	{}

	explicit CsBinaryExpression ( LineInfo line_info )
		: CsExpression ( n_binary_expression, line_info )
		, lhs(0)
		, oper(tkEOF)
		, rhs(0)
		, operator_token(0)
	{}

	CsBinaryExpression ( LineInfo line_info, CsNode * parent, CsExpression * lhs_, CsTokenType oper_ )
		: CsExpression ( n_binary_expression, line_info, parent )
		, lhs(lhs_)
		, oper(oper_)
		, rhs(0)
		, operator_token(0)
	{}

	CsBinaryExpression ( LineInfo line_info, CsExpression * lhs_, CsTokenType oper_, CsExpression * rhs_ )
		: CsExpression  ( n_binary_expression, line_info )
		, lhs(lhs_)
		, oper(oper_)
		, rhs(rhs_)
		, operator_token(0)
	{}

	CsBinaryExpression (
			LineInfo line_info
		, CsExpression * lhs_
		, CsTokenType oper_
		, CsExpression * rhs_
		, CsTokenIndex operator_token_
		)
		: CsExpression  ( n_binary_expression, line_info )
		, lhs(lhs_)
		, oper(oper_)
		, rhs(rhs_)
		, operator_token(operator_token_)
	{}

	virtual ~CsBinaryExpression()
	{
		delete lhs;
		delete rhs;
	}

	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
		if ( lhs )
			lhs->visit ( visitor );
		if ( rhs )
			rhs->visit ( visitor );
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

public:
	CsExpression * lhs;
	CsTokenType oper;
	CsExpression * rhs;
	CsTokenIndex operator_token;
};

DEFINE_AST_ALLOCATOR(CsBinaryExpression)

//------------------------------------------------------------------------------
// CsAsIsExpression
//------------------------------------------------------------------------------

struct CsAsIsExpression : public CsExpression
{
	CsAsIsExpression()
		: CsExpression ( n_as_is_expression )
		, expression(0)
		, oper(tkEOF)
		, type(0)
		, as_is_token(0)
	{}

	CsAsIsExpression ( CsExpression * expression_, CsTokenType oper_, CsTypeRef * type_ )
		: CsExpression ( n_as_is_expression, line_info )
		, expression(expression_)
		, oper(oper_)
		, type(type_)
		, as_is_token(0)
	{}

	explicit CsAsIsExpression ( LineInfo line_info )
		: CsExpression ( n_as_is_expression, line_info )
		, expression(0)
		, oper(tkEOF)
		, type(0)
		, as_is_token(0)
	{}
	
	CsAsIsExpression ( LineInfo line_info, CsExpression * expression_, CsTokenType oper_ )
		: CsExpression  ( n_as_is_expression, line_info )
		, expression(expression_)
		, oper(oper_)
		, type(0)
		, as_is_token(0)
	{}

	virtual ~CsAsIsExpression()
	{
		delete expression;
		delete type;
	}

	virtual void build_entities ( CsDisplay & display ) {}
	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void addEntityDefinitions ( CsEntityCollector & tokenCollector ) {}
	virtual void addEntityReferences ( CsEntityCollector & tokenCollector ) {}
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
		if ( expression )
			expression->visit ( visitor );
		if ( type )
			type->visit ( visitor );
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

public:
	CsExpression * expression;
	CsTokenType oper;
	CsTypeRef * type;
	CsTokenIndex as_is_token;
};

DEFINE_AST_ALLOCATOR(CsAsIsExpression)

//------------------------------------------------------------------------------
// CsConditionalExpression
//------------------------------------------------------------------------------

struct CsConditionalExpression : public CsExpression
{
	CsConditionalExpression()
		: CsExpression ( n_conditional_expression, line_info )
		, condition(0)
		, true_expression(0)
		, false_expression(0)
		, question_token(0)
		, colon_token(0)
	{}

	explicit CsConditionalExpression ( LineInfo line_info )
		: CsExpression  ( n_conditional_expression, line_info )
		, condition(0)
		, true_expression(0)
		, false_expression(0)
		, question_token(0)
		, colon_token(0)
	{}

	CsConditionalExpression ( LineInfo line_info, CsNode * parent, CsExpression * condition_ )
		: CsExpression  ( n_conditional_expression, line_info, parent )
		, condition(condition_)
		, true_expression(0)
		, false_expression(0)
		, question_token(0)
		, colon_token(0)
	{}

	virtual ~CsConditionalExpression()
	{
		delete condition;
		delete true_expression;
		delete false_expression;
	}

	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
		if ( condition )
			condition->visit ( visitor );
		if ( true_expression )
			true_expression->visit ( visitor );
		if ( false_expression )
			false_expression->visit ( visitor );
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

public:
	CsExpression * condition;
	CsExpression * true_expression;
	CsExpression * false_expression;
	CsTokenIndex question_token;
	CsTokenIndex colon_token;
};

DEFINE_AST_ALLOCATOR(CsConditionalExpression)

//------------------------------------------------------------------------------
// CsOperatorUnaryExpression
//------------------------------------------------------------------------------

struct CsOperatorUnaryExpression : public CsUnaryExpression
{
	CsOperatorUnaryExpression()
		: CsUnaryExpression ( n_operator_unary_expression, line_info )
		, operator_token(0)
	{}

	explicit CsOperatorUnaryExpression ( LineInfo line_info )
		: CsUnaryExpression ( n_operator_unary_expression, line_info )
		, operator_token(0)
	{}

	virtual ~CsOperatorUnaryExpression()
	{}

	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
		if ( unary_expression )
			unary_expression->visit ( visitor );
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	CsTokenType oper;
	CsTokenIndex operator_token;
};

DEFINE_AST_ALLOCATOR(CsOperatorUnaryExpression)

//------------------------------------------------------------------------------
// CsPreIncrementDecrementExpression
//------------------------------------------------------------------------------

struct CsPreIncrementDecrementExpression : public CsUnaryExpression
{
	CsPreIncrementDecrementExpression()
		: CsUnaryExpression ( n_pre_increment_decrement_expression )
		, operator_token(0)
	{}

	explicit CsPreIncrementDecrementExpression ( LineInfo line_info )
		: CsUnaryExpression ( n_pre_increment_decrement_expression, line_info )
		, operator_token(0)
	{}

	virtual ~CsPreIncrementDecrementExpression()
	{}

	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
		if ( unary_expression )
			unary_expression->visit ( visitor );
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	CsTokenType oper;
	CsTokenIndex operator_token;
};

DEFINE_AST_ALLOCATOR(CsPreIncrementDecrementExpression)

//------------------------------------------------------------------------------
// CsCastUnaryExpression
//------------------------------------------------------------------------------

struct CsCastUnaryExpression : public CsUnaryExpression
{
	CsCastUnaryExpression()
		: CsUnaryExpression ( n_cast_unary_expression, line_info )
		, type(0)
		, lparen_token(0)
		, rparen_token(0)
	{}

	explicit CsCastUnaryExpression ( LineInfo line_info )
		: CsUnaryExpression ( n_cast_unary_expression, line_info )
		, type(0)
		, lparen_token(0)
		, rparen_token(0)
	{}

	virtual ~CsCastUnaryExpression()
	{
		delete type;
	}

	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void addEntityDefinitions ( CsEntityCollector & tokenCollector );
	virtual void addEntityReferences ( CsEntityCollector & tokenCollector );
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
		if ( type )
			type->visit ( visitor );
		if ( unary_expression )
			unary_expression->visit ( visitor );
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	CsTypeRef * type;
	CsTokenIndex lparen_token;
	CsTokenIndex rparen_token;
};

DEFINE_AST_ALLOCATOR(CsCastUnaryExpression)

//------------------------------------------------------------------------------
// CsAssignmentExpression
//------------------------------------------------------------------------------

struct CsAssignmentExpression : public CsExpression
{
	CsAssignmentExpression()
		: CsExpression ( n_assignment_expression )
		, lhs(0)
		, oper(tkEOF)
		, rhs(0)
		, operator_token(0)
	{}

	explicit CsAssignmentExpression ( LineInfo line_info )
		: CsExpression ( n_assignment_expression, line_info )
		, lhs(0)
		, oper(tkEOF)
		, rhs(0)
		, operator_token(0)
	{}

	CsAssignmentExpression ( LineInfo line_info, CsNode * parent, CsExpression * lhs_, CsTokenType oper_
		, CsTokenIndex operator_token_
		)
		: CsExpression ( n_assignment_expression, line_info, parent )
		, lhs(lhs_)
		, oper(oper_)
		, rhs(0)
		, operator_token(operator_token_)
	{}

	CsAssignmentExpression ( LineInfo line_info, CsExpression * lhs_, CsTokenType oper_, CsExpression * rhs_
		, CsTokenIndex operator_token_
		)
		: CsExpression ( n_assignment_expression, line_info )
		, lhs(lhs_)
		, oper(oper_)
		, rhs(rhs_)
		, operator_token(operator_token_)
	{}

	virtual ~CsAssignmentExpression()
	{
		delete lhs;
		delete rhs;
	}

	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void addEntityDefinitions ( CsEntityCollector & tokenCollector ) {}
	virtual void addEntityReferences ( CsEntityCollector & tokenCollector ) {}
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
		if ( lhs )
			lhs->visit ( visitor );
		if ( rhs )
			rhs->visit ( visitor );
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

public:
	CsExpression * lhs;
	CsTokenType oper;
	CsExpression * rhs;
	CsTokenIndex operator_token;
};

DEFINE_AST_ALLOCATOR(CsAssignmentExpression)


//------------------------------------------------------------------------------
// CsExpressionList
//------------------------------------------------------------------------------

struct CsExpressionList : public CsNode
{
	CsExpressionList()
		: CsNode ( n_expression_list )
	{}

	explicit CsExpressionList ( LineInfo line_info )
		: CsNode ( n_expression_list, line_info )
	{}

	virtual ~CsExpressionList()
	{
		list.delete_nodes_and_values();
	}

	virtual void build_entities ( CsDisplay & display ) {}
	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void addEntityDefinitions ( CsEntityCollector & tokenCollector ) {}
	virtual void addEntityReferences ( CsEntityCollector & tokenCollector ) {}
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
		const single_list1<CsExpression*>::node * next = & list.first_;
		while ( next )
		{
			if ( next->value_ )
				next->value_->visit ( visitor );
			next = next->next_;
		}
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	single_list1<CsExpression*> list;
};

DEFINE_AST_ALLOCATOR(CsExpressionList)

//------------------------------------------------------------------------------
// AttributeSpecifierEnum
//------------------------------------------------------------------------------

enum AttributeSpecifierEnum
{
	asNONE = 0x0000,
	asASSEMBLY = 0x0001,
	asEVENT = 0x0002,
	asFIELD = 0x0003,
	asMETHOD = 0x0004,
	asPARAM = 0x0005,
	asPROPERTY = 0x0006,
	asRETURN = 0x0007,
	asTYPE = 0x0008,
};

//------------------------------------------------------------------------------
// CsNamedArgument
//------------------------------------------------------------------------------

struct CsNamedArgument : public CsNode
{
	CsNamedArgument()
		: CsNode ( n_attribute_named_argument )
		, expression(0)
		, entity(0)
		, separator_token(0)
		, identifier_token(0)
		, assign_token(0)
	{}

	CsNamedArgument ( CsNode * parent, LineInfo line_info )
		: CsNode ( n_attribute_named_argument, line_info, parent )
		, expression(0)
		, entity(0)
		, separator_token(0)
		, identifier_token(0)
		, assign_token(0)
	{}

	virtual ~CsNamedArgument()
	{
		delete expression;
	}

	virtual void build_entities ( CsDisplay & display ) {}
	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void addEntityDefinitions ( CsEntityCollector & tokenCollector );
	virtual void addEntityReferences ( CsEntityCollector & tokenCollector );
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
		if ( expression )
			expression->visit ( visitor );
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	PHashString identifier;
	CsExpression * expression;
	CsEntity * entity;
	CsTokenIndex separator_token;
	CsTokenIndex identifier_token;
	CsTokenIndex assign_token;
};

DEFINE_AST_ALLOCATOR(CsNamedArgument)

//------------------------------------------------------------------------------
// CsPositionalArgumentList
//------------------------------------------------------------------------------

struct CsPositionalArgumentList
	: public single_list1<CsExpression*>
{
	CsPositionalArgumentList()
		: single_list1<CsExpression*>()
	{}

	CsPositionalArgumentList ( CsExpression * expr )
		: single_list1<CsExpression*> ( expr )
	{}
};

//------------------------------------------------------------------------------
// CsNamedArgumentList
//------------------------------------------------------------------------------

struct CsNamedArgumentList : public single_list1<CsNamedArgument*>
{
	CsNamedArgumentList()
		: single_list1<CsNamedArgument*>()
	{}

	explicit CsNamedArgumentList ( CsNamedArgument * arg )
		: single_list1<CsNamedArgument*> ( arg )
	{}
};

//------------------------------------------------------------------------------
// CsAttribute
//------------------------------------------------------------------------------

struct CsAttribute : public CsNode
{
	CsAttribute()
		: CsNode ( n_attribute )
		, positional_argument_list(0)
		, named_argument_list(0)
		, entity(0)
		, separator_token(0)
		, lparen_token(0)
		, rparen_token(0)
	{}


	CsAttribute ( CsNode * parent, LineInfo line_info, WPHashString phs, CsTokenIndex identifier_token )
		: CsNode ( n_attribute, line_info, parent )
		, compound_identifier ( phs, identifier_token )
		, positional_argument_list(0)
		, named_argument_list(0)
		, entity(0)
		, separator_token(0)
		, lparen_token(0)
		, rparen_token(0)
	{}


	virtual ~CsAttribute()
	{
		if ( positional_argument_list )
		{
			positional_argument_list->delete_nodes_and_values();
			delete positional_argument_list;
		}
		if ( named_argument_list )
		{
			named_argument_list->delete_nodes_and_values();
			delete named_argument_list;
		}
	}

	virtual void build_entities ( CsDisplay & display ) {}
	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void addEntityDefinitions ( CsEntityCollector & tokenCollector );
	virtual void addEntityReferences ( CsEntityCollector & tokenCollector );
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
		if ( positional_argument_list )
		{
			const single_list1<CsExpression*>::node * next
				= & positional_argument_list->first_;
			while ( next )
			{
				if ( next->value_ )
					next->value_->visit ( visitor );
				next = next->next_;
			}
		}
		if ( named_argument_list )
		{
			const single_list1<CsNamedArgument*>::node * next
				= & named_argument_list->first_;
			while ( next )
			{
				if ( next->value_ )
					next->value_->visit ( visitor );
				next = next->next_;
			}
		}
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	CsCompoundIdentifier compound_identifier;
	CsPositionalArgumentList * positional_argument_list;
	CsNamedArgumentList * named_argument_list;
	CsEntity * entity;
	CsTokenIndex separator_token;
	CsTokenIndex lparen_token;
	CsTokenIndex rparen_token;
};

DEFINE_AST_ALLOCATOR(CsAttribute)

//------------------------------------------------------------------------------
// CsAttributeSection
//------------------------------------------------------------------------------

struct CsAttributeSection : public CsNode
{
	CsAttributeSection()
		: CsNode ( n_attribute_section )
		, specifier ( asNONE )
		, lbracket_token(0)
		, specifier_token(0)
		, colon_token(0)
		, rbracket_token(0)
	{}

	CsAttributeSection ( CsNode * parent, LineInfo line_info )
		: CsNode ( n_attribute_section, line_info, parent )
		, specifier ( asNONE )
		, lbracket_token(0)
		, specifier_token(0)
		, colon_token(0)
		, rbracket_token(0)
	{}

	virtual ~CsAttributeSection()
	{
		attribute_list.delete_nodes_and_values();
	}

	virtual void build_entities ( CsDisplay & display ) {}
	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void addEntityDefinitions ( CsEntityCollector & tokenCollector ) {}
	virtual void addEntityReferences ( CsEntityCollector & tokenCollector ) {}
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
		const single_list1<CsAttribute*>::node * next
			= & attribute_list.first_;
		while ( next )
		{
			if ( next->value_ )
				next->value_->visit ( visitor );
			next = next->next_;
		}
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	AttributeSpecifierEnum specifier;
	single_list1<CsAttribute*> attribute_list;
	CsTokenIndex lbracket_token;
	CsTokenIndex specifier_token;
	CsTokenIndex colon_token;
	CsTokenIndex rbracket_token;
};

DEFINE_AST_ALLOCATOR(CsAttributeSection)

//------------------------------------------------------------------------------
// CsAttributes
//------------------------------------------------------------------------------

struct CsAttributes : public CsNode
{
	CsAttributes()
		: CsNode ( n_attributes )
	{}

	explicit CsAttributes ( LineInfo line_info )
		: CsNode ( n_attributes, line_info )
	{}

	virtual ~CsAttributes()
	{
		sections.delete_nodes_and_values();
	}

	virtual void build_entities ( CsDisplay & display ) {}
	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void addEntityDefinitions ( CsEntityCollector & tokenCollector ) {}
	virtual void addEntityReferences ( CsEntityCollector & tokenCollector ) {}
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
		const single_list1<CsAttributeSection*>::node * next = & sections.first_;
		while ( next )
		{
			if ( next->value_ )
				next->value_->visit ( visitor );
			next = next->next_;
		}
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	single_list1<CsAttributeSection*> sections;
};

DEFINE_AST_ALLOCATOR(CsAttributes)

//------------------------------------------------------------------------------
// CsTypeBase
//------------------------------------------------------------------------------

struct CsTypeBase : public CsNode
{
	CsTypeBase()
		: CsNode ( n_type_base )
		, colon_token(0)
	{}

	explicit CsTypeBase ( LineInfo line_info )
		: CsNode ( n_type_base, line_info )
		, colon_token(0)
	{}

	virtual ~CsTypeBase()
	{
		base_list.delete_nodes_and_values();
	}

	virtual void build_entities ( CsDisplay & display ) {}
	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display ) {}

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void addEntityDefinitions ( CsEntityCollector & tokenCollector ) {}
	virtual void addEntityReferences ( CsEntityCollector & tokenCollector ) {}
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
		const single_list1<CsTypeRef*>::node * next
			= & base_list.first_;
		while ( next )
		{
			if ( next->value_ )
				next->value_->visit ( visitor );
			next = next->next_;
		}
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	single_list1<CsTypeRef*> base_list;
	CsTokenIndex colon_token;
};

DEFINE_AST_ALLOCATOR(CsTypeBase)



//------------------------------------------------------------------------------
// CsModifierEnum
//------------------------------------------------------------------------------

enum CsModifierEnum
{
// access specifiers occupy low 4 bits - this is need to fast evaluate access to entity
	mINTERNAL = 0x0001,
	mPRIVATE = 0x0002,
	mPROTECTED = 0x0004,
	mPUBLIC = 0x0008,

	m_non_access_modifiers_mask = 0x3ff0,

	mABSTRACT = 0x0010,
	mEXTERN = 0x0020,
	mNEW = 0x0040,
	mOVERRIDE = 0x0080,
	mREADONLY = 0x0100,
	mSEALED = 0x0200,
	mSTATIC = 0x0400,
	mVIRTUAL = 0x0800,
	mVOLATILE = 0x1000,
	mUNSAFE = 0x2000, // unsafe extension

// flags 0x00ff0000 used for cs_entity_access
};

//------------------------------------------------------------------------------
// CsModifiers
//------------------------------------------------------------------------------

struct CsModifiers
{
	CsModifiers()
		: flags(0)
	{}

	explicit CsModifiers ( unsigned modifiers_ )
		: flags(modifiers_)
	{}

	inline cs_entity_access getEntityAccess() const
	{
		return (cs_entity_access) (flags >> 16);
	}

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );

	unsigned flags;
	single_list1<CsTokenIndex> modifier_tokens;
};

//------------------------------------------------------------------------------
// CsClassStruct
//------------------------------------------------------------------------------

struct CsClassStruct : public CsScopeNode
{
	explicit CsClassStruct ( cs_node type )
		: CsScopeNode  ( type )
		, attributes(0)
		, base(0)
		, bBaseClassResolved(false)
		, ctor_method(0)
	{}

	CsClassStruct ( cs_node type, LineInfo line_info )
		: CsScopeNode  ( type, line_info )
		, attributes(0)
		, base(0)
		, bBaseClassResolved(false)
		, ctor_method(0)
	{}

	virtual ~CsClassStruct()
	{

		delete attributes;
		modifiers.modifier_tokens.delete_nodes();
		delete base;
		FastDeque<CsNode*>::iterator it = member_declarations.begin();
		FastDeque<CsNode*>::iterator it_end = member_declarations.end();
		for ( ; it != it_end; ++it )
		{
			delete (*it);
		}
	}

	virtual void build_entities ( CsDisplay & display );
	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );



	virtual void visit ( CsAstVisitor * visitor )
	{
		if ( attributes )
			attributes->visit ( visitor );
		if ( base )
			base->visit ( visitor );
		FastDeque<CsNode*>::iterator it = member_declarations.begin();
		FastDeque<CsNode*>::iterator it_end = member_declarations.end();
		for ( ; it != it_end; ++it )
		{
			if ( *it )
				(*it)->visit ( visitor );
		}
	}

	virtual void addEntityDefinitions ( CsEntityCollector & tokenCollector );
	virtual void addEntityReferences ( CsEntityCollector & tokenCollector );
	virtual void clearEntities();

	CsAttributes * attributes;
	CsModifiers modifiers;
	PHashString name;
	CsTypeBase * base;
	FastDeque<CsNode*> member_declarations;
	bool bBaseClassResolved;
	CsEntityMethod * ctor_method;
};

//------------------------------------------------------------------------------
// CsClass
//------------------------------------------------------------------------------

struct CsClass : public CsClassStruct
{
	CsClass()
		: CsClassStruct ( n_class, line_info )
		, class_token(0)
		, identifier_token(0)
		, lcurly_token(0)
		, rcurly_token(0)
		, semicolon_token(0)
	{}

	explicit CsClass ( LineInfo line_info )
		: CsClassStruct ( n_class, line_info )
		, class_token(0)
		, identifier_token(0)
		, lcurly_token(0)
		, rcurly_token(0)
		, semicolon_token(0)
	{}

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	//virtual void addEntityDefinitions ( CsEntityCollector & tokenCollector );
	//virtual void addEntityReferences ( CsEntityCollector & tokenCollector );
	virtual void clearEntities();

	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
		CsClassStruct::visit ( visitor );
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	CsTokenIndex class_token;
	CsTokenIndex identifier_token;
	CsTokenIndex lcurly_token;
	CsTokenIndex rcurly_token;
	CsTokenIndex semicolon_token;
};

DEFINE_AST_ALLOCATOR(CsClass)

//------------------------------------------------------------------------------
// CsStruct
//------------------------------------------------------------------------------

struct CsStruct : public CsClassStruct
{
	CsStruct()
		: CsClassStruct ( n_struct )
		, struct_token(0)
		, identifier_token(0)
		, lcurly_token(0)
		, rcurly_token(0)
		, semicolon_token(0)
	{}

	explicit CsStruct ( LineInfo line_info )
		: CsClassStruct ( n_struct, line_info )
		, struct_token(0)
		, identifier_token(0)
		, lcurly_token(0)
		, rcurly_token(0)
		, semicolon_token(0)
	{}

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	//virtual void addEntityDefinitions ( CsEntityCollector & tokenCollector );
	//virtual void addEntityReferences ( CsEntityCollector & tokenCollector );
	virtual void clearEntities();

	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
		CsClassStruct::visit ( visitor );
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	CsTokenIndex struct_token;
	CsTokenIndex identifier_token;
	CsTokenIndex lcurly_token;
	CsTokenIndex rcurly_token;
	CsTokenIndex semicolon_token;
};

DEFINE_AST_ALLOCATOR(CsStruct)

//------------------------------------------------------------------------------
// CsInterface
//------------------------------------------------------------------------------

struct CsInterface : public CsScopeNode
{
	CsInterface()
		: CsScopeNode ( n_interface )
		, attributes(0)
		, base(0)
		, bBaseClassResolved(false)
		, interface_token(0)
		, identifier_token(0)
		, lcurly_token(0)
		, rcurly_token(0)
		, semicolon_token(0)
	{}

	explicit CsInterface ( LineInfo line_info )
		: CsScopeNode ( n_interface, line_info )
		, attributes(0)
		, base(0)
		, bBaseClassResolved(false)
		, interface_token(0)
		, identifier_token(0)
		, lcurly_token(0)
		, rcurly_token(0)
		, semicolon_token(0)
	{}

	virtual ~CsInterface()
	{

		delete attributes;
		modifiers.modifier_tokens.delete_nodes();
		delete base;
		DELETE_FROM_CONTAINER(FastDeque<CsNode*>,member_declarations);
	}

	virtual void build_entities ( CsDisplay & display );
	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void addEntityDefinitions ( CsEntityCollector & tokenCollector );
	virtual void addEntityReferences ( CsEntityCollector & tokenCollector );
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
		if ( attributes )
			attributes->visit ( visitor );
		if ( base )
			base->visit ( visitor );
		FastDeque<CsNode*>::iterator it = member_declarations.begin();
		FastDeque<CsNode*>::iterator it_end = member_declarations.end();
		for ( ; it != it_end; ++it )
		{
			if ( *it )
				(*it)->visit ( visitor );
		}
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	CsAttributes * attributes;
	CsModifiers modifiers;
	PHashString name;
	CsTypeBase * base;
	FastDeque<CsNode*> member_declarations;
	bool bBaseClassResolved;
	CsTokenIndex interface_token;
	CsTokenIndex identifier_token;
	CsTokenIndex lcurly_token;
	CsTokenIndex rcurly_token;
	CsTokenIndex semicolon_token;
};

DEFINE_AST_ALLOCATOR(CsInterface)

//------------------------------------------------------------------------------
// CsEnumMember
//------------------------------------------------------------------------------

struct CsEnumMember : public CsNode
{
	CsEnumMember()
		: CsNode ( n_enum_member )
		, attributes(0)
		, expression(0)
		, entity(0)
		, identifier_token(0)
		, assign_token(0)
		, comma_token(0)
	{}

	CsEnumMember ( CsNode * parent, LineInfo line_info )
		: CsNode ( n_enum_member, line_info, parent )
		, attributes(0)
		, expression(0)
		, entity(0)
		, identifier_token(0)
		, assign_token(0)
		, comma_token(0)
	{}

	virtual ~CsEnumMember()
	{
		delete attributes;
		delete expression;
	}

	virtual void build_entities ( CsDisplay & display );
	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void addEntityDefinitions ( CsEntityCollector & tokenCollector );
	virtual void addEntityReferences ( CsEntityCollector & tokenCollector );
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
		if ( attributes )
			attributes->visit ( visitor );
		if ( expression )
			expression->visit ( visitor );
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	CsAttributes * attributes;
	PHashString name;
	CsExpression * expression;
	CsEntityConstant * entity;
	CsTokenIndex identifier_token;
	CsTokenIndex assign_token;
	CsTokenIndex comma_token;
};

DEFINE_AST_ALLOCATOR(CsEnumMember)

//------------------------------------------------------------------------------
// CsEnum
//------------------------------------------------------------------------------

struct CsEnum : public CsScopeNode
{
	CsEnum()
		: CsScopeNode ( n_enum )
		, attributes(0)
		, base(tkEOF)
		, enum_token(0)
		, identifier_token(0)
		, colon_token(0)
		, base_token(0)
		, lcurly_token(0)
		, rcurly_token(0)
		, semicolon_token(0)
	{}

	explicit CsEnum ( LineInfo line_info )
		: CsScopeNode ( n_enum, line_info )
		, attributes(0)
		, base(tkEOF)
		, enum_token(0)
		, identifier_token(0)
		, colon_token(0)
		, base_token(0)
		, lcurly_token(0)
		, rcurly_token(0)
		, semicolon_token(0)
	{}

	virtual ~CsEnum()
	{
		delete attributes;
		modifiers.modifier_tokens.delete_nodes();
		DELETE_FROM_CONTAINER(FastArray<CsEnumMember*>,members);
	}

	virtual void build_entities ( CsDisplay & display );
	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void addEntityDefinitions ( CsEntityCollector & tokenCollector );
	virtual void addEntityReferences ( CsEntityCollector & tokenCollector );
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
		if ( attributes )
			attributes->visit ( visitor );
		FastArray<CsEnumMember*>::iterator it = members.begin();
		FastArray<CsEnumMember*>::iterator it_end = members.end();
		for ( ; it != it_end; ++it )
		{
			if ( *it )
				(*it)->visit ( visitor );
		}
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	CsAttributes * attributes;
	CsModifiers modifiers;
	PHashString name;
	CsTokenType base;
	FastArray<CsEnumMember*> members;
	CsTokenIndex enum_token;
	CsTokenIndex identifier_token;
	CsTokenIndex colon_token;
	CsTokenIndex base_token;
	CsTokenIndex lcurly_token;
	CsTokenIndex rcurly_token;
	CsTokenIndex semicolon_token;
};

DEFINE_AST_ALLOCATOR(CsEnum)

//------------------------------------------------------------------------------
// CsFormalParameter
//------------------------------------------------------------------------------

struct CsFormalParameter : public CsNode
{
	CsFormalParameter()
		: CsNode ( n_formal_parameter )
		, attributes(0)
		, modifier(tkEOF)
		, type(0)
		, entity(0)
		, separator_token(0)
		, modifier_token(0)
		, identifier_token(0)
	{}

	explicit CsFormalParameter ( CsNode * parent )
		: CsNode ( n_formal_parameter, parent )
		, attributes(0)
		, modifier(tkEOF)
		, type(0)
		, entity(0)
		, separator_token(0)
		, modifier_token(0)
		, identifier_token(0)
	{}

	CsFormalParameter ( CsNode * parent, LineInfo line_info )
		: CsNode ( n_formal_parameter, line_info, parent )
		, attributes(0)
		, modifier(tkEOF)
		, type(0)
		, entity(0)
		, separator_token(0)
		, modifier_token(0)
		, identifier_token(0)
	{}

	virtual ~CsFormalParameter()
	{
		delete attributes;
		delete type;
	}

	virtual void build_entities ( CsDisplay & display );
	virtual void build_entities ( CsDisplay & display, bool set_reference_from_ast_to_entity );
	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void addEntityDefinitions ( CsEntityCollector & tokenCollector );
	virtual void addEntityReferences ( CsEntityCollector & tokenCollector );
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
		if ( attributes )
			attributes->visit ( visitor );
		if ( type )
			type->visit ( visitor );
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	CsAttributes * attributes;
	CsTokenType modifier; // ref, out, __arglist
	CsTypeRef * type;
	PHashString name;
	CsEntityFormalParameter * entity;
	CsTokenIndex separator_token;
	CsTokenIndex modifier_token;
	CsTokenIndex identifier_token;
};

DEFINE_AST_ALLOCATOR(CsFormalParameter)

//------------------------------------------------------------------------------
// CsFormalParameterList
//------------------------------------------------------------------------------

struct CsFormalParameterList : public CsNode
{
	CsFormalParameterList()
		: CsNode ( n_formal_parameter_list, line_info )
	{}

	explicit CsFormalParameterList ( LineInfo line_info )
		: CsNode ( n_formal_parameter_list, line_info )
	{}

	virtual ~CsFormalParameterList()
	{
		parameters.delete_nodes_and_values();
	}

	virtual void build_entities ( CsDisplay & display );
	virtual void build_entities ( CsDisplay & display, bool set_reference_from_ast_to_entity );
	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void addEntityDefinitions ( CsEntityCollector & tokenCollector ) {}
	virtual void addEntityReferences ( CsEntityCollector & tokenCollector ) {}
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
		const single_list1<CsFormalParameter*>::node * next
			= & parameters.first_;
		while ( next )
		{
			if ( next->value_ )
				next->value_->visit ( visitor );
			next = next->next_;
		}
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	single_list1<CsFormalParameter*> parameters;
};

DEFINE_AST_ALLOCATOR(CsFormalParameterList)

//------------------------------------------------------------------------------
// CsAttributesModifiers
//------------------------------------------------------------------------------

struct CsAttributesModifiers
{
	CsAttributesModifiers()
		: attributes(0)
	{}

	virtual ~CsAttributesModifiers()
	{
		delete attributes;
		modifiers.modifier_tokens.delete_nodes();
	}

	CsAttributes * attributes;
	CsModifiers modifiers;
};

//------------------------------------------------------------------------------
// CsDelegate
//------------------------------------------------------------------------------

struct CsDelegate : public CsScopeNode, public CsAttributesModifiers
{
	CsDelegate()
		: CsScopeNode ( n_delegate )
		, type(0)
		, parameters(0)
		, ctor_method(0)
		, invoke_method(0)
		, beginInvoke_method(0)
		, endInvoke_method(0)
		, delegate_token(0)
		, identifier_token(0)
		, lparen_token(0)
		, rparen_token(0)
		, semicolon_token(0)
	{}

	explicit CsDelegate ( LineInfo line_info )
		: CsScopeNode ( n_delegate, line_info )
		, type(0)
		, parameters(0)
		, ctor_method(0)
		, invoke_method(0)
		, beginInvoke_method(0)
		, endInvoke_method(0)
		, delegate_token(0)
		, identifier_token(0)
		, lparen_token(0)
		, rparen_token(0)
		, semicolon_token(0)
	{}

	virtual ~CsDelegate()
	{

		delete type;
		delete parameters;
	}

	virtual void build_entities ( CsDisplay & display );
	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display ); // ?

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void addEntityDefinitions ( CsEntityCollector & tokenCollector );
	virtual void addEntityReferences ( CsEntityCollector & tokenCollector );
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
		if ( attributes )
			attributes->visit ( visitor );
		if ( type )
			type->visit ( visitor );
		if ( parameters )
			parameters->visit ( visitor );
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	CsTypeRef * type;
	PHashString name;
	CsFormalParameterList * parameters;
	CsEntityMethod * ctor_method;
	CsEntityMethod * invoke_method;
	CsEntityMethod * beginInvoke_method;
	CsEntityMethod * endInvoke_method;
	CsTokenIndex delegate_token;
	CsTokenIndex identifier_token;
	CsTokenIndex lparen_token;
	CsTokenIndex rparen_token;
	CsTokenIndex semicolon_token;
};

DEFINE_AST_ALLOCATOR(CsDelegate)

//------------------------------------------------------------------------------
// CsNamespace
//------------------------------------------------------------------------------

struct CsNamespace : public CsScopeNodeWithUsingDirectives
{
	CsNamespace()
		: CsScopeNodeWithUsingDirectives ( n_namespace )
		, attributes(0)
		, namespace_token(0)
		, lcurly_token(0)
		, rcurly_token(0)
		, semicolon_token(0)
	{}

	explicit CsNamespace ( LineInfo line_info )
		: CsScopeNodeWithUsingDirectives ( n_namespace, line_info )
		, attributes(0)
		, namespace_token(0)
		, lcurly_token(0)
		, rcurly_token(0)
		, semicolon_token(0)
	{}

	virtual ~CsNamespace()
	{
		delete attributes;
		modifiers.modifier_tokens.delete_nodes();
		//DELETE_FROM_CONTAINER(std::vector<CsNode*>,member_declarations);
		DELETE_FROM_CONTAINER(FastArray<CsNode*>,member_declarations);
	}

	virtual void build_entities ( CsDisplay & display );
	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void addEntityDefinitions ( CsEntityCollector & tokenCollector );
	virtual void addEntityReferences ( CsEntityCollector & tokenCollector );
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
		if ( attributes )
			attributes->visit ( visitor );
		CsScopeNodeWithUsingDirectives::visit ( visitor );
		FastArray<CsNode*>::iterator it = member_declarations.begin();
		FastArray<CsNode*>::iterator it_end = member_declarations.end();
		for ( ; it != it_end; ++it )
		{
			if ( *it )
				(*it)->visit ( visitor );
		}
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	CsAttributes * attributes;
	CsModifiers modifiers;
	CsCompoundIdentifier compound_identifier;
	FastArray<CsNode*> member_declarations;
	CsTokenIndex namespace_token;
	CsTokenIndex lcurly_token;
	CsTokenIndex rcurly_token;
	CsTokenIndex semicolon_token;
};

DEFINE_AST_ALLOCATOR(CsNamespace)

//------------------------------------------------------------------------------
// CsArgument
//------------------------------------------------------------------------------

struct CsArgument : public CsNode
{
	CsArgument()
		: CsNode ( n_argument )
		, modifier(tkEOF)
		, expression(0)
		, separator_token(0)
		, modifier_token(0)
	{}

	explicit CsArgument ( CsNode * parent )
		: CsNode ( n_argument, parent )
		, modifier(tkEOF)
		, expression(0)
		, separator_token(0)
		, modifier_token(0)
	{}

	CsArgument ( CsNode * parent, LineInfo line_info )
		: CsNode ( n_argument, line_info, parent )
		, modifier(tkEOF)
		, expression(0)
		, separator_token(0)
		, modifier_token(0)
	{}

	virtual ~CsArgument()
	{
		delete expression;
	}

	virtual void build_entities ( CsDisplay & display ) {}
	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void addEntityDefinitions ( CsEntityCollector & tokenCollector ) {}
	virtual void addEntityReferences ( CsEntityCollector & tokenCollector ) {}
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
		if ( expression )
			expression->visit ( visitor );
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	CsTokenType modifier;
	CsExpression * expression;
	CsTokenIndex separator_token;
	CsTokenIndex modifier_token;
};

DEFINE_AST_ALLOCATOR(CsArgument)

//------------------------------------------------------------------------------
// CsArgumentList
//------------------------------------------------------------------------------

struct CsArgumentList : public CsNode
{
	CsArgumentList()
		: CsNode ( n_argument_list, line_info )
	{}

	explicit CsArgumentList ( LineInfo line_info )
		: CsNode ( n_argument_list, line_info )
	{}

	virtual ~CsArgumentList()
	{
		list.delete_nodes_and_values();
	}

	virtual void build_entities ( CsDisplay & display ) {}
	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void addEntityDefinitions ( CsEntityCollector & tokenCollector ) {}
	virtual void addEntityReferences ( CsEntityCollector & tokenCollector ) {}
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
		const single_list1<CsArgument*>::node * next
			= & list.first_;
		while ( next )
		{
			if ( next->value_ )
				next->value_->visit ( visitor );
			next = next->next_;
		}
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	single_list1<CsArgument*> list;
};

DEFINE_AST_ALLOCATOR(CsArgumentList)

//------------------------------------------------------------------------------
// CsStatement
//------------------------------------------------------------------------------

// abstract
struct CsStatement : public CsNode
{
	explicit CsStatement ( cs_node type )
		: CsNode ( type )
	{}

	CsStatement ( cs_node type, LineInfo line_info )
		: CsNode ( type, line_info )
	{}

	virtual ~CsStatement()
	{}
};

//------------------------------------------------------------------------------
// CsBlock
//------------------------------------------------------------------------------

struct CsBlock : public CsStatement
{
	CsBlock()
		: CsStatement ( n_block )
		, entity ( 0 )
		, lcurly_token(0)
		, rcurly_token(0)
	{}

	explicit CsBlock ( LineInfo line_info )
		: CsStatement ( n_block, line_info )
		, entity ( 0 )
		, lcurly_token(0)
		, rcurly_token(0)
	{}

	virtual ~CsBlock()
	{
		DELETE_FROM_CONTAINER(FastArray<CsStatement*>,statements);
	}

	virtual void build_entities ( CsDisplay & display );
	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void addEntityDefinitions ( CsEntityCollector & tokenCollector );
	virtual void addEntityReferences ( CsEntityCollector & tokenCollector );
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
		FastArray<CsStatement*>::iterator it = statements.begin();
		FastArray<CsStatement*>::iterator it_end = statements.end();
		for ( ; it != it_end; ++it )
		{
			if ( *it )
				(*it)->visit ( visitor );
		}
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	FastArray<CsStatement*> statements;
	CsScopeEntity * entity;
	CsTokenIndex lcurly_token;
	CsTokenIndex rcurly_token;
};

DEFINE_AST_ALLOCATOR(CsBlock)


//------------------------------------------------------------------------------
// CsLabel
//------------------------------------------------------------------------------

struct CsLabel : public CsNode
{
	CsLabel()
		: CsNode ( n_label )
		, identifier_token(0)
		, colon_token(0)
	{}

	CsLabel ( WPHashString phs, LineInfo line_info )
		: CsNode ( n_label, line_info )
		, identifier ( phs )
		, identifier_token(0)
		, colon_token(0)
	{}

	virtual ~CsLabel()
	{}

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void addEntityDefinitions ( CsEntityCollector & tokenCollector );
	virtual void addEntityReferences ( CsEntityCollector & tokenCollector ) {}
	virtual void clearEntities() {}

	virtual void build_entities ( CsDisplay & display );
	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );


	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	PHashString identifier;
	CsTokenIndex identifier_token;
	CsTokenIndex colon_token;
};

DEFINE_AST_ALLOCATOR(CsLabel)

//------------------------------------------------------------------------------
// CsLabeledStatement
//------------------------------------------------------------------------------

struct CsLabeledStatement : public CsStatement
{
	CsLabeledStatement()
		: CsStatement ( n_labeled_statement )
		, labels(0)
		, statement(0)
	{}

	CsLabeledStatement ( CsStatement * stat, LineInfo line_info )
		: CsStatement ( n_labeled_statement, line_info )
		, labels(0)
		, statement(stat)
	{}

	virtual ~CsLabeledStatement()
	{
		if ( labels )
		{
			labels->delete_nodes_and_values();
			delete labels;
		}
		delete statement;
	}

	virtual void build_entities ( CsDisplay & display );
	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void addEntityDefinitions ( CsEntityCollector & tokenCollector ) {}
	virtual void addEntityReferences ( CsEntityCollector & tokenCollector ) {}
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
		if ( labels )
		{
			single_list1<CsLabel*>::node * next
				= & labels->first_;
			while ( next )
			{
				if ( next->value_ )
					next->value_->visit ( visitor );
				next = next->next_;
			}
		}
		if ( statement )
			statement->visit ( visitor );
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	single_list1<CsLabel*> * labels;
	CsStatement * statement;
};

DEFINE_AST_ALLOCATOR(CsLabeledStatement)

//------------------------------------------------------------------------------
// CsEmptyStatement
//------------------------------------------------------------------------------

struct CsEmptyStatement : public CsStatement
{
	CsEmptyStatement()
		: CsStatement ( n_empty_statement, line_info )
		, semicolon_token(0)
	{}

	explicit CsEmptyStatement ( LineInfo line_info )
		: CsStatement ( n_empty_statement, line_info )
		, semicolon_token(0)
	{}

	virtual ~CsEmptyStatement()
	{}

	virtual void build_entities ( CsDisplay & display ) {}
	virtual void type_lookup ( CsDisplay & display ) {}
	virtual void member_lookup ( CsDisplay & display ) {}

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void addEntityDefinitions ( CsEntityCollector & tokenCollector ) {}
	virtual void addEntityReferences ( CsEntityCollector & tokenCollector ) {}
	virtual void clearEntities() {}

	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	CsTokenIndex semicolon_token;
};

DEFINE_AST_ALLOCATOR(CsEmptyStatement)

//------------------------------------------------------------------------------
// CsExpressionStatement
//------------------------------------------------------------------------------

struct CsExpressionStatement : public CsStatement
{
	CsExpressionStatement()
		: CsStatement ( n_expression_statement, line_info )
		, expression(0)
		, semicolon_token(0)
	{}

	explicit CsExpressionStatement ( LineInfo line_info )
		: CsStatement ( n_expression_statement, line_info )
		, expression(0)
		, semicolon_token(0)
	{}

	CsExpressionStatement ( LineInfo line_info, CsExpression * expr )
		: CsStatement ( n_expression_statement, line_info )
		, expression(expr)
		, semicolon_token(0)
	{}

	virtual ~CsExpressionStatement()
	{
		delete expression;
	}

	virtual void build_entities ( CsDisplay & display ) {}
	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );
	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void addEntityDefinitions ( CsEntityCollector & tokenCollector ) {}
	virtual void addEntityReferences ( CsEntityCollector & tokenCollector ) {}
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
		if ( expression )
			expression->visit ( visitor );
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	CsExpression * expression;
	CsTokenIndex semicolon_token;
};

DEFINE_AST_ALLOCATOR(CsExpressionStatement)

//------------------------------------------------------------------------------
// CsDeclarationStatement
//------------------------------------------------------------------------------

struct CsDeclarationStatement : public CsStatement
{
	explicit CsDeclarationStatement()
		: CsStatement ( n_declaration_statement )
		, node(0)
		, semicolon_token(0)
	{}

	virtual ~CsDeclarationStatement()
	{
		delete node;
	}

	virtual void build_entities ( CsDisplay & display );
	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void addEntityDefinitions ( CsEntityCollector & tokenCollector ) {}
	virtual void addEntityReferences ( CsEntityCollector & tokenCollector ) {}
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
		if ( node )
			node->visit ( visitor );
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	CsNode * node; // local-variable-declaration or local-constant-declaration ;
	CsTokenIndex semicolon_token;
};

DEFINE_AST_ALLOCATOR(CsDeclarationStatement)

//------------------------------------------------------------------------------
// CsLocalConstantDeclarator
//------------------------------------------------------------------------------

struct CsLocalConstantDeclarator : public CsNode
{
	CsLocalConstantDeclarator()
		: CsNode ( n_local_constant_declarator )
		, expression(0)
		, entity(0)
		, separator_token(0)
		, identifier_token(0)
		, assign_token(0)
	{}

	CsLocalConstantDeclarator ( CsNode * parent, LineInfo line_info )
		: CsNode ( n_local_constant_declarator, line_info, parent )
		, expression(0)
		, entity(0)
		, separator_token(0)
		, identifier_token(0)
		, assign_token(0)
	{}

	virtual ~CsLocalConstantDeclarator()
	{
		delete expression;
	}

	virtual void build_entities ( CsDisplay & display );
	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void addEntityDefinitions ( CsEntityCollector & tokenCollector );
	virtual void addEntityReferences ( CsEntityCollector & tokenCollector );
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
		if ( expression )
			expression->visit ( visitor );
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	PHashString name;
	CsExpression * expression;
	CsEntityLocalConstant * entity;
	CsTokenIndex separator_token;
	CsTokenIndex identifier_token;
	CsTokenIndex assign_token;
};

DEFINE_AST_ALLOCATOR(CsLocalConstantDeclarator)

//------------------------------------------------------------------------------
// CsLocalConstantDeclaration
//------------------------------------------------------------------------------

struct CsLocalConstantDeclaration : public CsStatement
{
	CsLocalConstantDeclaration()
		: CsStatement ( n_local_constant_declaration, line_info )
		, type(0)
		, const_token(0)
		, semicolon_token(0)
	{}

	explicit CsLocalConstantDeclaration ( LineInfo line_info )
		: CsStatement ( n_local_constant_declaration, line_info )
		, type(0)
		, const_token(0)
		, semicolon_token(0)
	{}

	virtual ~CsLocalConstantDeclaration()
	{
		delete type;
		declarators.delete_nodes_and_values();
	}

	virtual void build_entities ( CsDisplay & display );
	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void addEntityDefinitions ( CsEntityCollector & tokenCollector ) {}
	virtual void addEntityReferences ( CsEntityCollector & tokenCollector ) {}
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
		if ( type )
			type->visit ( visitor );
		const single_list1<CsLocalConstantDeclarator*>::node * next
			= & declarators.first_;
		while ( next )
		{
			if ( next->value_ )
				next->value_->visit ( visitor );
			next = next->next_;
		}
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	CsTypeRef * type;
	single_list1<CsLocalConstantDeclarator*> declarators;
	CsTokenIndex const_token;
	CsTokenIndex semicolon_token;
};

DEFINE_AST_ALLOCATOR(CsLocalConstantDeclaration)

//------------------------------------------------------------------------------
// CsIfStatement
//------------------------------------------------------------------------------

struct CsIfStatement : public CsStatement
{
	CsIfStatement()
		: CsStatement ( n_if_statement, line_info )
		, condition(0)
		, if_statement(0)
		, else_statement(0)
		, if_token(0)
		, lparen_token(0)
		, rparen_token(0)
		, else_token(0)
	{}

	explicit CsIfStatement ( LineInfo line_info )
		: CsStatement ( n_if_statement, line_info )
		, condition(0)
		, if_statement(0)
		, else_statement(0)
		, if_token(0)
		, lparen_token(0)
		, rparen_token(0)
		, else_token(0)
	{}

	virtual ~CsIfStatement()
	{
		delete condition;
		delete if_statement;
		delete else_statement;
	}

	virtual void build_entities ( CsDisplay & display );
	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void addEntityDefinitions ( CsEntityCollector & tokenCollector ) {}
	virtual void addEntityReferences ( CsEntityCollector & tokenCollector ) {}
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
		if ( condition )
			condition->visit ( visitor );
		if ( if_statement )
			if_statement->visit ( visitor );
		if ( else_statement )
			else_statement->visit ( visitor );
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	CsExpression * condition;
	CsStatement * if_statement;
	CsStatement * else_statement;
	CsTokenIndex if_token;
	CsTokenIndex lparen_token;
	CsTokenIndex rparen_token;
	CsTokenIndex else_token;
};

DEFINE_AST_ALLOCATOR(CsIfStatement)

//------------------------------------------------------------------------------
// CsSwitchLabel
//------------------------------------------------------------------------------

struct CsSwitchLabel : public CsNode
{
	CsSwitchLabel()
		: CsNode ( n_switch_label )
		, bDefault(false)
		, expression(0)
		, case_default_token(0)
		, colon_token(0)
	{}

	CsSwitchLabel ( CsNode * parent, LineInfo line_info )
		: CsNode ( n_switch_label, line_info, parent )
		, bDefault(false)
		, expression(0)
		, case_default_token(0)
		, colon_token(0)
	{}

	virtual ~CsSwitchLabel()
	{
		delete expression;
	}

	virtual void build_entities ( CsDisplay & display ) {}
	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void addEntityDefinitions ( CsEntityCollector & tokenCollector ) {}
	virtual void addEntityReferences ( CsEntityCollector & tokenCollector ) {}
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
		if ( expression )
			expression->visit ( visitor );
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	bool bDefault; // false - case, true - default
	CsExpression * expression;
	CsTokenIndex case_default_token;
	CsTokenIndex colon_token;
};

DEFINE_AST_ALLOCATOR(CsSwitchLabel)

//------------------------------------------------------------------------------
// CsSwitchSection
//------------------------------------------------------------------------------

struct CsSwitchSection : public CsNode
{
	CsSwitchSection()
		: CsNode ( n_switch_section )
		, labels(0)
		, entity(0)
	{}

	CsSwitchSection ( CsNode * parent, LineInfo line_info )
		: CsNode ( n_switch_section, line_info, parent )
		, labels(0)
		, entity(0)
	{}

	virtual ~CsSwitchSection()
	{
		if ( labels )
		{
			labels->delete_nodes_and_values();
			delete labels;
		}
		DELETE_FROM_CONTAINER(FastArray<CsStatement*>,statements);
	}

	virtual void build_entities ( CsDisplay & display );
	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void addEntityDefinitions ( CsEntityCollector & tokenCollector ) {}
	virtual void addEntityReferences ( CsEntityCollector & tokenCollector ) {}
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
		if ( labels )
		{
			single_list1<CsSwitchLabel*>::node * next = & labels->first_;
			while ( next )
			{
				if ( next->value_ )
					next->value_->visit ( visitor );
				next = next->next_;
			}
		}
		FastArray<CsStatement*>::iterator it = statements.begin();
		FastArray<CsStatement*>::iterator it_end = statements.end();
		for ( ; it != it_end; ++it )
		{
			if ( *it )
				(*it)->visit ( visitor );
		}
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	single_list1<CsSwitchLabel*> * labels;
	FastArray<CsStatement*> statements;
	CsEntityBlock * entity;
};

DEFINE_AST_ALLOCATOR(CsSwitchSection)

//------------------------------------------------------------------------------
// CsSwitchStatement
//------------------------------------------------------------------------------

struct CsSwitchStatement : public CsStatement
{
	CsSwitchStatement()
		: CsStatement ( n_switch_statement )
		, expression(0)
		, entity(0)
		, governing_type(0)
		, switch_token(0)
		, lparen_token(0)
		, rparen_token(0)
		, lcurly_token(0)
		, rcurly_token(0)
	{}

	explicit CsSwitchStatement ( LineInfo line_info )
		: CsStatement ( n_switch_statement, line_info )
		, expression(0)
		, entity(0)
		, governing_type(0)
		, switch_token(0)
		, lparen_token(0)
		, rparen_token(0)
		, lcurly_token(0)
		, rcurly_token(0)
	{}

	virtual ~CsSwitchStatement()
	{

		delete expression;
		DELETE_FROM_CONTAINER(FastArray<CsSwitchSection*>,sections);
	}

	virtual void build_entities ( CsDisplay & display );
	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void addEntityDefinitions ( CsEntityCollector & tokenCollector );
	virtual void addEntityReferences ( CsEntityCollector & tokenCollector );
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
		if ( expression )
			expression->visit ( visitor );
		FastArray<CsSwitchSection*>::iterator it = sections.begin();
		FastArray<CsSwitchSection*>::iterator it_end = sections.end();
		for ( ; it != it_end; ++it )
		{
			if ( *it )
				(*it)->visit ( visitor );
		}
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	CsExpression * expression;
	FastArray<CsSwitchSection*> sections; // FastDeque is faster (give 2-3% speed up for AST generation), but FastArray more useful for programmers
	CsEntityBlock * entity;
	CsEntityTypeRef * governing_type;
	CsTokenIndex switch_token;
	CsTokenIndex lparen_token;
	CsTokenIndex rparen_token;
	CsTokenIndex lcurly_token;
	CsTokenIndex rcurly_token;
};

DEFINE_AST_ALLOCATOR(CsSwitchStatement)

//------------------------------------------------------------------------------
// CsWhileStatement
//------------------------------------------------------------------------------

struct CsWhileStatement : public CsStatement
{
	CsWhileStatement()
		: CsStatement ( n_while_statement )
		, condition(0)
		, statement(0)
		, while_token(0)
		, lparen_token(0)
		, rparen_token(0)
	{}

	explicit CsWhileStatement ( LineInfo line_info )
		: CsStatement ( n_while_statement, line_info )
		, condition(0)
		, statement(0)
		, while_token(0)
		, lparen_token(0)
		, rparen_token(0)
	{}

	virtual ~CsWhileStatement()
	{
		delete condition;
		delete statement;
	}

	virtual void build_entities ( CsDisplay & display );
	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void addEntityDefinitions ( CsEntityCollector & tokenCollector ) {}
	virtual void addEntityReferences ( CsEntityCollector & tokenCollector ) {}
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
		if ( condition )
			condition->visit ( visitor );
		if ( statement )
			statement->visit ( visitor );
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	CsExpression * condition;
	CsStatement * statement;
	CsTokenIndex while_token;
	CsTokenIndex lparen_token;
	CsTokenIndex rparen_token;
};

DEFINE_AST_ALLOCATOR(CsWhileStatement)

//------------------------------------------------------------------------------
// CsDoWhileStatement
//------------------------------------------------------------------------------

struct CsDoWhileStatement : public CsStatement
{
	CsDoWhileStatement()
		: CsStatement ( n_do_while_statement )
		, condition(0)
		, statement(0)
		, do_token(0)
		, while_token(0)
		, lparen_token(0)
		, rparen_token(0)
		, semicolon_token(0)
	{}

	explicit CsDoWhileStatement ( LineInfo line_info )
		: CsStatement ( n_do_while_statement, line_info )
		, condition(0)
		, statement(0)
		, do_token(0)
		, while_token(0)
		, lparen_token(0)
		, rparen_token(0)
		, semicolon_token(0)
	{}

	virtual ~CsDoWhileStatement()
	{
		delete statement;
		delete condition;
	}

	virtual void build_entities ( CsDisplay & display );
	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void addEntityDefinitions ( CsEntityCollector & tokenCollector ) {}
	virtual void addEntityReferences ( CsEntityCollector & tokenCollector ) {}
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
		if ( statement )
			statement->visit ( visitor );
		if ( condition )
			condition->visit ( visitor );
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	CsStatement * statement;
	CsExpression * condition;
	CsTokenIndex do_token;
	CsTokenIndex while_token;
	CsTokenIndex lparen_token;
	CsTokenIndex rparen_token;
	CsTokenIndex semicolon_token;
};

DEFINE_AST_ALLOCATOR(CsDoWhileStatement)

//------------------------------------------------------------------------------
// CsLocalVariableDeclarator
//------------------------------------------------------------------------------

struct CsLocalVariableDeclarator : public CsNode
{
	CsLocalVariableDeclarator()
		: CsNode ( n_local_variable_declarator )
		, expression(0)
		, entity(0)
		, separator_token(0)
		, identifier_token(0)
		, assign_token(0)
	{}

	CsLocalVariableDeclarator ( CsNode * parent, LineInfo line_info )
		: CsNode ( n_local_variable_declarator, line_info, parent )
		, expression(0)
		, entity(0)
		, separator_token(0)
		, identifier_token(0)
		, assign_token(0)
	{}

	virtual ~CsLocalVariableDeclarator()
	{
		delete expression;
	}

	virtual void build_entities ( CsDisplay & display );
	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void addEntityDefinitions ( CsEntityCollector & tokenCollector );
	virtual void addEntityReferences ( CsEntityCollector & tokenCollector );
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
		if ( expression )
			expression->visit ( visitor );
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	PHashString name;
	CsExpression * expression;
	CsEntityLocalVariable * entity;
	CsTokenIndex separator_token;
	CsTokenIndex identifier_token;
	CsTokenIndex assign_token;
};

DEFINE_AST_ALLOCATOR(CsLocalVariableDeclarator)

//------------------------------------------------------------------------------
// CsLocalVariableDeclaration
//------------------------------------------------------------------------------

struct CsLocalVariableDeclaration : public CsStatement
{
	CsLocalVariableDeclaration()
		: CsStatement ( n_local_variable_declaration )
		, type(0)
	{}

	virtual ~CsLocalVariableDeclaration()
	{
		delete type;
		declarators.delete_nodes_and_values();
	}

	virtual void build_entities ( CsDisplay & display );
	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void addEntityDefinitions ( CsEntityCollector & tokenCollector ) {}
	virtual void addEntityReferences ( CsEntityCollector & tokenCollector ) {}
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
		if ( type )
			type->visit ( visitor );
		const single_list1<CsLocalVariableDeclarator*>::node * next
			= & declarators.first_;
		while ( next )
		{
			if ( next->value_ )
				next->value_->visit ( visitor );
			next = next->next_;
		}
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	CsTypeRef * type;
	single_list1<CsLocalVariableDeclarator*> declarators;
};

DEFINE_AST_ALLOCATOR(CsLocalVariableDeclaration)

//------------------------------------------------------------------------------
// CsStatementExpressionList
//------------------------------------------------------------------------------

struct CsStatementExpressionList : public CsNode
{
	CsStatementExpressionList()
		: CsNode ( n_statement_expression_list )
	{}

	explicit CsStatementExpressionList ( LineInfo line_info )
		: CsNode ( n_statement_expression_list, line_info )
	{}

	virtual ~CsStatementExpressionList()
	{
		expressions.delete_nodes_and_values();
	}

	virtual void build_entities ( CsDisplay & display ) {}
	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void addEntityDefinitions ( CsEntityCollector & tokenCollector ) {}
	virtual void addEntityReferences ( CsEntityCollector & tokenCollector ) {}
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
		const single_list1<CsExpression*>::node * next
			= & expressions.first_;
		while ( next )
		{
			if ( next->value_ )
				next->value_->visit ( visitor );
			next = next->next_;
		}
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	single_list1<CsExpression*> expressions;
};

DEFINE_AST_ALLOCATOR(CsStatementExpressionList)

//------------------------------------------------------------------------------
// CsForStatement
//------------------------------------------------------------------------------

struct CsForStatement : public CsStatement
{
	CsForStatement()
		: CsStatement ( n_for_statement )
		, initializer(0)
		, condition(0)
		, iterator(0)
		, statement(0)
		, entity(0)
		, for_token(0)
		, lparen_token(0)
		, semicolon1_token(0)
		, semicolon2_token(0)
		, rparen_token(0)
	{}

	explicit CsForStatement ( LineInfo line_info )
		: CsStatement ( n_for_statement, line_info )
		, initializer(0)
		, condition(0)
		, iterator(0)
		, statement(0)
		, entity(0)
		, for_token(0)
		, lparen_token(0)
		, semicolon1_token(0)
		, semicolon2_token(0)
		, rparen_token(0)
	{}

	virtual ~CsForStatement()
	{
		delete initializer;
		delete condition;
		delete iterator;
		delete statement;
	}

	virtual void build_entities ( CsDisplay & display );
	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void addEntityDefinitions ( CsEntityCollector & tokenCollector );
	virtual void addEntityReferences ( CsEntityCollector & tokenCollector );
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
		if ( initializer )
			initializer->visit ( visitor );
		if ( condition )
			condition->visit ( visitor );
		if ( iterator )
			iterator->visit ( visitor );
		if ( statement )
			statement->visit ( visitor );
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	CsNode * initializer;
	CsExpression * condition;
	CsNode * iterator;
	CsStatement * statement;
	CsEntityBlock * entity;
	CsTokenIndex for_token;
	CsTokenIndex lparen_token;
	CsTokenIndex semicolon1_token;
	CsTokenIndex semicolon2_token;
	CsTokenIndex rparen_token;
};

DEFINE_AST_ALLOCATOR(CsForStatement)

//------------------------------------------------------------------------------
// CsForeachStatement
//------------------------------------------------------------------------------

struct CsForeachStatement : public CsStatement
{
	CsForeachStatement()
		: CsStatement ( n_foreach_statement )
		, type(0)
		, expression(0)
		, statement(0)
		, entity(0)
		, foreach_token(0)
		, lparen_token(0)
		, identifier_token(0)
		, in_token(0)
		, rparen_token(0)
	{}

	explicit CsForeachStatement ( LineInfo line_info )
		: CsStatement ( n_foreach_statement, line_info )
		, type(0)
		, expression(0)
		, statement(0)
		, entity(0)
		, foreach_token(0)
		, lparen_token(0)
		, identifier_token(0)
		, in_token(0)
		, rparen_token(0)
	{}

	virtual ~CsForeachStatement()
	{
		delete type;
		delete expression;
		delete statement;
	}

	virtual void build_entities ( CsDisplay & display );
	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void addEntityDefinitions ( CsEntityCollector & tokenCollector );
	virtual void addEntityReferences ( CsEntityCollector & tokenCollector );
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
		if ( type )
			type->visit ( visitor );
		if ( expression )
			expression->visit ( visitor );
		if ( statement )
			statement->visit ( visitor );
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	CsTypeRef * type;
	PHashString identifier;
	CsExpression * expression;
	CsStatement * statement;
	CsEntityBlockVariable * entity;
	CsTokenIndex foreach_token;
	CsTokenIndex lparen_token;
	CsTokenIndex identifier_token;
	CsTokenIndex in_token;
	CsTokenIndex rparen_token;
};

DEFINE_AST_ALLOCATOR(CsForeachStatement)

//------------------------------------------------------------------------------
// CsBreakStatement
//------------------------------------------------------------------------------

struct CsBreakStatement : public CsStatement
{
	CsBreakStatement()
		: CsStatement ( n_break_statement )
		, break_token(0)
		, semicolon_token(0)
	{}

	explicit CsBreakStatement ( LineInfo line_info )
		: CsStatement ( n_break_statement, line_info )
		, break_token(0)
		, semicolon_token(0)
	{}

	virtual ~CsBreakStatement()
	{}

	virtual void build_entities ( CsDisplay & display ) {}
	virtual void type_lookup ( CsDisplay & display ) {}
	virtual void member_lookup ( CsDisplay & display ) {}

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void addEntityDefinitions ( CsEntityCollector & tokenCollector ) {}
	virtual void addEntityReferences ( CsEntityCollector & tokenCollector ) {}
	virtual void clearEntities() {}

	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	CsTokenIndex break_token;
	CsTokenIndex semicolon_token;
};

DEFINE_AST_ALLOCATOR(CsBreakStatement)

//------------------------------------------------------------------------------
// CsContinueStatement
//------------------------------------------------------------------------------

struct CsContinueStatement : public CsStatement
{
	CsContinueStatement()
		: CsStatement ( n_continue_statement )
		, continue_token(0)
		, semicolon_token(0)
	{}

	explicit CsContinueStatement ( LineInfo line_info )
		: CsStatement ( n_continue_statement, line_info )
		, continue_token(0)
		, semicolon_token(0)
	{}

	virtual ~CsContinueStatement()
	{}

	virtual void build_entities ( CsDisplay & display ) {}
	virtual void type_lookup ( CsDisplay & display ) {}
	virtual void member_lookup ( CsDisplay & display ) {}

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void addEntityDefinitions ( CsEntityCollector & tokenCollector ) {}
	virtual void addEntityReferences ( CsEntityCollector & tokenCollector ) {}
	virtual void clearEntities() {}

	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	CsTokenIndex continue_token;
	CsTokenIndex semicolon_token;
};

DEFINE_AST_ALLOCATOR(CsContinueStatement)

//------------------------------------------------------------------------------
// CsGotoStatement
//------------------------------------------------------------------------------

struct CsGotoStatement : public CsStatement
{
	CsGotoStatement()
		: CsStatement ( n_goto_statement )
		, label_type(tkEOF)
		, case_expression(0)
		, goto_target(0)
		, goto_token(0)
		, label_token(0)
		, default_case_token(0)
		, semicolon_token(0)
	{}

	explicit CsGotoStatement ( LineInfo line_info )
		: CsStatement ( n_goto_statement, line_info )
		, label_type(tkEOF)
		, case_expression(0)
		, goto_target(0)
		, goto_token(0)
		, label_token(0)
		, default_case_token(0)
		, semicolon_token(0)
	{}

	virtual ~CsGotoStatement()
	{
		delete case_expression;
	}

	virtual void build_entities ( CsDisplay & display ) {}
	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void addEntityDefinitions ( CsEntityCollector & tokenCollector ) {}
	virtual void addEntityReferences ( CsEntityCollector & tokenCollector ) {}
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
		if ( case_expression )
			case_expression->visit ( visitor );
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	CsTokenType label_type;
	PHashString label;
	CsExpression * case_expression;
	CsNode * goto_target; // not a child
	CsTokenIndex goto_token;
	CsTokenIndex label_token;
	CsTokenIndex default_case_token;
	CsTokenIndex semicolon_token;
};

DEFINE_AST_ALLOCATOR(CsGotoStatement)

//------------------------------------------------------------------------------
// CsReturnStatement
//------------------------------------------------------------------------------

struct CsReturnStatement : public CsStatement
{
	CsReturnStatement()
		: CsStatement ( n_return_statement )
		, expression(0)
		, return_token(0)
		, semicolon_token(0)
	{}

	explicit CsReturnStatement ( LineInfo line_info )
		: CsStatement ( n_return_statement, line_info )
		, expression(0)
		, return_token(0)
		, semicolon_token(0)
	{}

	virtual ~CsReturnStatement()
	{
		delete expression;
	}

	virtual void build_entities ( CsDisplay & display ) {}
	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void addEntityDefinitions ( CsEntityCollector & tokenCollector ) {}
	virtual void addEntityReferences ( CsEntityCollector & tokenCollector ) {}
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
		if ( expression )
			expression->visit ( visitor );
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	CsExpression * expression;
	CsTokenIndex return_token;
	CsTokenIndex semicolon_token;
};

DEFINE_AST_ALLOCATOR(CsReturnStatement)

//------------------------------------------------------------------------------
// CsThrowStatement
//------------------------------------------------------------------------------

struct CsThrowStatement : public CsStatement
{
	CsThrowStatement()
		: CsStatement ( n_throw_statement )
		, expression(0)
		, throw_token(0)
		, semicolon_token(0)
	{}

	explicit CsThrowStatement ( LineInfo line_info )
		: CsStatement ( n_throw_statement, line_info )
		, expression(0)
		, throw_token(0)
		, semicolon_token(0)
	{}

	virtual ~CsThrowStatement()
	{
		delete expression;
	}

	virtual void build_entities ( CsDisplay & display ) {}
	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void addEntityDefinitions ( CsEntityCollector & tokenCollector ) {}
	virtual void addEntityReferences ( CsEntityCollector & tokenCollector ) {}
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
		if ( expression )
			expression->visit ( visitor );
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	CsExpression * expression;
	CsTokenIndex throw_token;
	CsTokenIndex semicolon_token;
};

DEFINE_AST_ALLOCATOR(CsThrowStatement)

//------------------------------------------------------------------------------
// CsCatchClause
//------------------------------------------------------------------------------

struct CsCatchClause : public CsNode
{
	CsCatchClause()
		: CsNode ( n_catch_clause )
		, type(0)
		, block(0)
		, entity(0)
		, catch_token(0)
		, lparen_token(0)
		, identifier_token(0)
		, rparen_token(0)
	{}

	CsCatchClause ( CsNode * parent, LineInfo line_info )
		: CsNode ( n_catch_clause, line_info, parent )
		, type(0)
		, block(0)
		, entity(0)
		, catch_token(0)
		, lparen_token(0)
		, identifier_token(0)
		, rparen_token(0)
	{}

	virtual ~CsCatchClause()
	{
		delete type;
		delete block;
	}

	virtual void build_entities ( CsDisplay & display );
	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void addEntityDefinitions ( CsEntityCollector & tokenCollector );
	virtual void addEntityReferences ( CsEntityCollector & tokenCollector );
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
		if ( type )
			type->visit ( visitor );
		if ( block )
			block->visit ( visitor );
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	CsTypeRef * type;
	PHashString identifier;
	CsBlock * block;
	CsEntityBlockVariable * entity;
	CsTokenIndex catch_token;
	CsTokenIndex lparen_token;
	CsTokenIndex identifier_token;
	CsTokenIndex rparen_token;
};

DEFINE_AST_ALLOCATOR(CsCatchClause)

//------------------------------------------------------------------------------
// CsTryStatement
//------------------------------------------------------------------------------

struct CsTryStatement : public CsStatement
{
	CsTryStatement()
		: CsStatement ( n_try_statement )
		, try_block(0)
		, finally_block(0)
		, entity(0)
		, try_token(0)
		, finally_token(0)
	{}

	explicit CsTryStatement ( LineInfo line_info )
		: CsStatement ( n_try_statement, line_info )
		, try_block(0)
		, finally_block(0)
		, entity(0)
		, try_token(0)
		, finally_token(0)
	{}

	virtual ~CsTryStatement()
	{
		delete try_block;
		delete finally_block;
		catch_clauses.delete_nodes_and_values();
	}

	virtual void build_entities ( CsDisplay & display );
	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void addEntityDefinitions ( CsEntityCollector & tokenCollector );
	virtual void addEntityReferences ( CsEntityCollector & tokenCollector );
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
		if ( try_block )
			try_block->visit ( visitor );
		const single_list1<CsCatchClause*>::node * next = & catch_clauses.first_;
		while ( next )
		{
			if ( next->value_ )
				next->value_->visit ( visitor );
			next = next->next_;
		}
		if ( finally_block )
			finally_block->visit ( visitor );
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	CsBlock * try_block;
	single_list1<CsCatchClause*> catch_clauses;
	CsBlock * finally_block;
	CsEntityBlock * entity; // stuff entity block, introduced for support catch identifiers
	CsTokenIndex try_token;
	CsTokenIndex finally_token;
};

DEFINE_AST_ALLOCATOR(CsTryStatement)

//------------------------------------------------------------------------------
// CsCheckedStatement
//------------------------------------------------------------------------------

struct CsCheckedStatement : public CsStatement
{
	CsCheckedStatement()
		: CsStatement ( n_checked_statement )
		, block(0)
		, checked_token(0)
	{}

	CsCheckedStatement ( LineInfo line_info )
		: CsStatement ( n_checked_statement, line_info )
		, block(0)
		, checked_token(0)
	{}

	virtual ~CsCheckedStatement()
	{
		delete block;
	}

	virtual void build_entities ( CsDisplay & display );
	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void addEntityDefinitions ( CsEntityCollector & tokenCollector ) {}
	virtual void addEntityReferences ( CsEntityCollector & tokenCollector ) {}
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
		if ( block )
			block->visit ( visitor );
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	CsBlock * block;
	CsTokenIndex checked_token;
};

DEFINE_AST_ALLOCATOR(CsCheckedStatement)

//------------------------------------------------------------------------------
// CsUncheckedStatement
//------------------------------------------------------------------------------

struct CsUncheckedStatement : public CsStatement
{
	CsUncheckedStatement()
		: CsStatement ( n_unchecked_statement )
		, block(0)
		, unchecked_token(0)
	{}

	CsUncheckedStatement ( LineInfo line_info )
		: CsStatement ( n_unchecked_statement, line_info )
		, block(0)
		, unchecked_token(0)
	{}

	virtual ~CsUncheckedStatement()
	{
		delete block;
	}

	virtual void build_entities ( CsDisplay & display );
	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void addEntityDefinitions ( CsEntityCollector & tokenCollector ) {}
	virtual void addEntityReferences ( CsEntityCollector & tokenCollector ) {}
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
		if ( block )
			block->visit ( visitor );
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	CsBlock * block;
	CsTokenIndex unchecked_token;
};

DEFINE_AST_ALLOCATOR(CsUncheckedStatement)

//------------------------------------------------------------------------------
// CsLockStatement
//------------------------------------------------------------------------------

struct CsLockStatement : public CsStatement
{
	CsLockStatement()
		: CsStatement ( n_lock_statement )
		, expression(0)
		, statement(0)
		, lock_token(0)
		, lparen_token(0)
		, rparen_token(0)
	{}

	CsLockStatement ( LineInfo line_info )
		: CsStatement ( n_lock_statement, line_info )
		, expression(0)
		, statement(0)
		, lock_token(0)
		, lparen_token(0)
		, rparen_token(0)
	{}

	virtual ~CsLockStatement()
	{
		delete expression;
		delete statement;
	}

	virtual void build_entities ( CsDisplay & display );
	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void addEntityDefinitions ( CsEntityCollector & tokenCollector ) {}
	virtual void addEntityReferences ( CsEntityCollector & tokenCollector ) {}
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
		if ( expression )
			expression->visit ( visitor );
		if ( statement )
			statement->visit ( visitor );
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	CsExpression * expression;
	CsStatement * statement;
	CsTokenIndex lock_token;
	CsTokenIndex lparen_token;
	CsTokenIndex rparen_token;
};

DEFINE_AST_ALLOCATOR(CsLockStatement)

//------------------------------------------------------------------------------
// CsUsingStatement
//------------------------------------------------------------------------------

struct CsUsingStatement : public CsStatement
{
	CsUsingStatement()
		: CsStatement ( n_using_statement )
		, resource(0)
		, statement(0)
		, entity(0)
		, using_token(0)
		, lparen_token(0)
		, rparen_token(0)
	{}

	explicit CsUsingStatement ( LineInfo line_info )
		: CsStatement ( n_using_statement, line_info )
		, resource(0)
		, statement(0)
		, entity(0)
		, using_token(0)
		, lparen_token(0)
		, rparen_token(0)
	{}

	virtual ~CsUsingStatement()
	{
		delete resource;
		delete statement;
	}

	virtual void build_entities ( CsDisplay & display );
	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void addEntityDefinitions ( CsEntityCollector & tokenCollector );
	virtual void addEntityReferences ( CsEntityCollector & tokenCollector );
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
		if ( resource )
			resource->visit ( visitor );
		if ( statement )
			statement->visit ( visitor );
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	CsNode * resource;
	CsStatement * statement;
	CsEntityBlock * entity;
	CsTokenIndex using_token;
	CsTokenIndex lparen_token;
	CsTokenIndex rparen_token;
};

DEFINE_AST_ALLOCATOR(CsUsingStatement)

//------------------------------------------------------------------------------
// CsUnsafeStatement
//------------------------------------------------------------------------------

struct CsUnsafeStatement : public CsStatement
{
	CsUnsafeStatement()
		: CsStatement ( n_unsafe_statement )
		, block(0)
		, unsafe_token(0)
	{}

	CsUnsafeStatement ( CsBlock * bl, LineInfo line_info )
		: CsStatement ( n_unsafe_statement, line_info )
		, block(bl)
		, unsafe_token(0)
	{}

	virtual ~CsUnsafeStatement()
	{
		delete block;
	}

	virtual void build_entities ( CsDisplay & display );
	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void addEntityDefinitions ( CsEntityCollector & tokenCollector ) {}
	virtual void addEntityReferences ( CsEntityCollector & tokenCollector ) {}
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
		if ( block )
			block->visit ( visitor );
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	CsBlock * block;
	CsTokenIndex unsafe_token;
};

DEFINE_AST_ALLOCATOR(CsUnsafeStatement)

//------------------------------------------------------------------------------
// CsFixedPointDeclarator
//------------------------------------------------------------------------------

struct CsFixedPointDeclarator : public CsNode
{
	CsFixedPointDeclarator()
		: CsNode ( n_fixed_point_declarator )
		, expression(0)
		, entity(0)
		, separator_token(0)
		, identifier_token(0)
		, assign_token(0)
	{}

	explicit CsFixedPointDeclarator ( LineInfo line_info )
		: CsNode ( n_fixed_point_declarator, line_info )
		, expression(0)
		, entity(0)
		, separator_token(0)
		, identifier_token(0)
		, assign_token(0)
	{}

	virtual ~CsFixedPointDeclarator()
	{
		delete expression;
	}

	virtual void build_entities ( CsDisplay & display );
	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void addEntityDefinitions ( CsEntityCollector & tokenCollector );
	virtual void addEntityReferences ( CsEntityCollector & tokenCollector );
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
		if ( expression )
			expression->visit ( visitor );
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	PHashString identifier;
	CsExpression * expression;
	CsEntityLocalVariable * entity;
	CsTokenIndex separator_token;
	CsTokenIndex identifier_token;
	CsTokenIndex assign_token;
};

DEFINE_AST_ALLOCATOR(CsFixedPointDeclarator)

//------------------------------------------------------------------------------
// CsFixedStatement
//------------------------------------------------------------------------------

struct CsFixedStatement : public CsStatement
{
	CsFixedStatement()
		: CsStatement ( n_fixed_statement )
		, type(0)
		, statement(0)
		, entity(0)
		, fixed_token(0)
		, lparen_token(0)
		, rparen_token(0)
	{}

	explicit CsFixedStatement ( LineInfo line_info )
		: CsStatement ( n_fixed_statement, line_info )
		, type(0)
		, statement(0)
		, entity(0)
		, fixed_token(0)
		, lparen_token(0)
		, rparen_token(0)
	{}

	virtual ~CsFixedStatement()
	{
		delete type;
		delete statement;
		declarators.delete_nodes_and_values();
	}

	virtual void build_entities ( CsDisplay & display );
	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void addEntityDefinitions ( CsEntityCollector & tokenCollector );
	virtual void addEntityReferences ( CsEntityCollector & tokenCollector );
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
		if ( type )
			type->visit ( visitor );
		const single_list1<CsFixedPointDeclarator*>::node * next
			= & declarators.first_;
		while ( next )
		{
			if ( next->value_ )
				next->value_->visit ( visitor );
			next = next->next_;
		}
		if ( statement )
			statement->visit ( visitor );
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	CsTypeRef * type;
	single_list1<CsFixedPointDeclarator*> declarators;
	CsStatement * statement;
	CsEntityBlock * entity;
	CsTokenIndex fixed_token;
	CsTokenIndex lparen_token;
	CsTokenIndex rparen_token;
};

DEFINE_AST_ALLOCATOR(CsFixedStatement)


//------------------------------------------------------------------------------
// CsStackAllocExpression
//------------------------------------------------------------------------------

struct CsStackAllocExpression : public CsExpression
{
	CsStackAllocExpression()
		: CsExpression ( n_stack_alloc_expression )
		, type(0)
		, expression (0)
		, stackalloc_token(0)
		, lbracket_token(0)
		, rbracket_token(0)
	{}

	explicit CsStackAllocExpression ( LineInfo line_info )
		: CsExpression  ( n_stack_alloc_expression, line_info )
		, type(0)
		, expression (0)
		, stackalloc_token(0)
		, lbracket_token(0)
		, rbracket_token(0)
	{}

	virtual ~CsStackAllocExpression()
	{
		delete type;
		delete expression;
	}

	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
		if ( type )
			type->visit ( visitor );
		if ( expression )
			expression->visit ( visitor );
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	CsTypeRef * type;
	CsExpression * expression;
	CsTokenIndex stackalloc_token;
	CsTokenIndex lbracket_token;
	CsTokenIndex rbracket_token;
};

DEFINE_AST_ALLOCATOR(CsStackAllocExpression)

//------------------------------------------------------------------------------
// CsConstructorDeclaration
//------------------------------------------------------------------------------

struct CsConstructorDeclaration : public CsNode, public CsAttributesModifiers
{
	CsConstructorDeclaration()
		: CsNode ( n_constructor_declaration )
		, parameters(0)
		, basethis(tkEOF)
		, argument_list(0)
		, definition(0)
		, entity(0)
		, identifier_token(0)
		, lparen_token(0)
		, rparen_token(0)
		, colon_token(0)
		, base_this_token(0)
		, base_this_lparen_token(0)
		, base_this_rparen_token(0)
		, semicolon_token(0)
	{}

	explicit CsConstructorDeclaration ( LineInfo line_info )
		: CsNode ( n_constructor_declaration, line_info )
		, parameters(0)
		, basethis(tkEOF)
		, argument_list(0)
		, definition(0)
		, entity(0)
		, identifier_token(0)
		, lparen_token(0)
		, rparen_token(0)
		, colon_token(0)
		, base_this_token(0)
		, base_this_lparen_token(0)
		, base_this_rparen_token(0)
		, semicolon_token(0)
	{}

	virtual ~CsConstructorDeclaration()
	{

		delete parameters;
		delete argument_list;
		delete definition;
	}

	virtual void build_entities ( CsDisplay & display );
	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void addEntityDefinitions ( CsEntityCollector & tokenCollector );
	virtual void addEntityReferences ( CsEntityCollector & tokenCollector );
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
		if ( attributes )
			attributes->visit ( visitor );
		if ( parameters )
			parameters->visit ( visitor );
		if ( argument_list )
			argument_list->visit ( visitor );
		if ( definition )
			definition->visit ( visitor );
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	PHashString name;
	CsFormalParameterList * parameters;
	CsTokenType basethis;
	CsArgumentList * argument_list;
	CsBlock * definition;
	CsEntityMethod * entity;
	CsTokenIndex identifier_token;
	CsTokenIndex lparen_token;
	CsTokenIndex rparen_token;
	CsTokenIndex colon_token;
	CsTokenIndex base_this_token;
	CsTokenIndex base_this_lparen_token;
	CsTokenIndex base_this_rparen_token;
	CsTokenIndex semicolon_token;
};

DEFINE_AST_ALLOCATOR(CsConstructorDeclaration)

//------------------------------------------------------------------------------
// CsConstantDeclarator
//------------------------------------------------------------------------------

struct CsConstantDeclarator : public CsNode
{
	CsConstantDeclarator()
		: CsNode ( n_constant_declarator )
		, expression(0)
		, entity(0)
		, separator_token(0)
		, identifier_token(0)
		, assign_token(0)
	{}

	CsConstantDeclarator ( CsNode * parent, LineInfo line_info )
		: CsNode ( n_constant_declarator, line_info, parent )
		, expression(0)
		, entity(0)
		, separator_token(0)
		, identifier_token(0)
		, assign_token(0)
	{}

	virtual ~CsConstantDeclarator()
	{
		delete expression;
	}

	virtual void build_entities ( CsDisplay & display );
	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void addEntityDefinitions ( CsEntityCollector & tokenCollector );
	virtual void addEntityReferences ( CsEntityCollector & tokenCollector );
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
		if ( expression )
			expression->visit ( visitor );
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	PHashString name;
	CsExpression * expression;
	CsEntityConstant * entity;
	CsTokenIndex separator_token;
	CsTokenIndex identifier_token;
	CsTokenIndex assign_token;
};

DEFINE_AST_ALLOCATOR(CsConstantDeclarator)

//------------------------------------------------------------------------------
// CsConstantDeclaration
//------------------------------------------------------------------------------

struct CsConstantDeclaration : public CsNode, public CsAttributesModifiers
{
	CsConstantDeclaration()
		: CsNode ( n_constant_declaration )
		, type(0)
		, const_token(0)
		, semicolon_token(0)
	{}

	CsConstantDeclaration ( LineInfo line_info )
		: CsNode ( n_constant_declaration, line_info )
		, type(0)
		, const_token(0)
		, semicolon_token(0)
	{}

	virtual ~CsConstantDeclaration()
	{
		delete type;
		declarators.delete_nodes_and_values();
	}

	virtual void build_entities ( CsDisplay & display );
	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void addEntityDefinitions ( CsEntityCollector & tokenCollector ) {}
	virtual void addEntityReferences ( CsEntityCollector & tokenCollector ) {}
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
		if ( attributes )
			attributes->visit ( visitor );
		if ( type )
			type->visit ( visitor );
		const single_list1<CsConstantDeclarator*>::node * next
			= & declarators.first_;
		while ( next )
		{
			if ( next->value_ )
				next->value_->visit ( visitor );
			next = next->next_;
		}
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	CsTypeRef * type;
	single_list1<CsConstantDeclarator*> declarators;
	CsTokenIndex const_token;
	CsTokenIndex semicolon_token;
};

DEFINE_AST_ALLOCATOR(CsConstantDeclaration)

//------------------------------------------------------------------------------
// CsVariableDeclarator
//------------------------------------------------------------------------------

struct CsVariableDeclarator : public CsNode
{
	CsVariableDeclarator()
		: CsNode ( n_variable_declarator )
		, expression(0)
		, entity(0)
		, separator_token(0)
		, identifier_token(0)
		, assign_token(0)
	{}

	CsVariableDeclarator ( CsNode * parent, LineInfo line_info )
		: CsNode ( n_variable_declarator, line_info, parent )
		, expression(0)
		, entity(0)
		, separator_token(0)
		, identifier_token(0)
		, assign_token(0)
	{}

	virtual ~CsVariableDeclarator()
	{
		delete expression;
	}

	virtual void build_entities ( CsDisplay & display );
	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void addEntityDefinitions ( CsEntityCollector & tokenCollector );
	virtual void addEntityReferences ( CsEntityCollector & tokenCollector );
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
		if ( expression )
			expression->visit ( visitor );
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	PHashString name;
	CsExpression * expression;
	CsEntityVariable * entity;
	CsTokenIndex separator_token;
	CsTokenIndex identifier_token;
	CsTokenIndex assign_token;
};

DEFINE_AST_ALLOCATOR(CsVariableDeclarator)

//------------------------------------------------------------------------------
// CsVariableDeclaration
//------------------------------------------------------------------------------

struct CsVariableDeclaration : public CsNode, public CsAttributesModifiers
{
	CsVariableDeclaration()
		: CsNode ( n_variable_declaration )
		, type(0)
		, semicolon_token(0)
	{}

	explicit CsVariableDeclaration ( LineInfo line_info )
		: CsNode ( n_variable_declaration, line_info )
		, type(0)
		, semicolon_token(0)
	{}

	virtual ~CsVariableDeclaration()
	{
		delete type;
		declarators.delete_nodes_and_values();
	}

	virtual void build_entities ( CsDisplay & display );
	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void addEntityDefinitions ( CsEntityCollector & tokenCollector ) {}
	virtual void addEntityReferences ( CsEntityCollector & tokenCollector ) {}
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
		if ( attributes )
			attributes->visit ( visitor );
		if ( type )
			type->visit ( visitor );
		const single_list1<CsVariableDeclarator*>::node * next
			= & declarators.first_;
		while ( next )
		{
			if ( next->value_ )
				next->value_->visit ( visitor );
			next = next->next_;
		}
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	CsTypeRef * type;
	single_list1<CsVariableDeclarator*> declarators;
	CsTokenIndex semicolon_token;
};

DEFINE_AST_ALLOCATOR(CsVariableDeclaration)

//------------------------------------------------------------------------------
// CsArrayInitializer
//------------------------------------------------------------------------------

struct CsArrayInitializer : public CsExpression
{
	CsArrayInitializer()
		: CsExpression ( n_array_initializer )
		, lcurly_token(0)
		, rcurly_token(0)
	{}

	explicit CsArrayInitializer ( LineInfo line_info )
		: CsExpression ( n_array_initializer, line_info )
		, lcurly_token(0)
		, rcurly_token(0)
	{}

	virtual ~CsArrayInitializer()
	{
		FastDeque<CsExpression*>::iterator it = expressions.begin();
		FastDeque<CsExpression*>::iterator it_end = expressions.end();
		for ( ; it != it_end; ++it )
		{
			delete (*it);
		}
	}

	virtual void build_entities ( CsDisplay & display ) {}
	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void addEntityDefinitions ( CsEntityCollector & tokenCollector ) {}
	virtual void addEntityReferences ( CsEntityCollector & tokenCollector ) {}
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
		FastDeque<CsExpression*>::iterator it = expressions.begin();
		FastDeque<CsExpression*>::iterator it_end = expressions.end();
		for ( ; it != it_end; ++it )
		{
			if ( *it )
				(*it)->visit ( visitor );
		}
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	FastDeque<CsExpression*> expressions;
	CsTokenIndex lcurly_token;
	CsTokenIndex rcurly_token;
};

DEFINE_AST_ALLOCATOR(CsArrayInitializer)

//------------------------------------------------------------------------------
// CsMethod
//------------------------------------------------------------------------------

struct CsMethod : public CsNode, public CsAttributesModifiers
{
	CsMethod()
		: CsNode ( n_method_declaration )
		, return_type(0)
		, parameters(0)
		, definition(0)
		, entity(0)
		, lparen_token(0)
		, rparen_token(0)
		, semicolon_token(0)
	{}

	explicit CsMethod ( LineInfo line_info )
		: CsNode ( n_method_declaration, line_info )
		, return_type(0)
		, parameters(0)
		, definition(0)
		, entity(0)
		, lparen_token(0)
		, rparen_token(0)
		, semicolon_token(0)
	{}

	virtual ~CsMethod()
	{

		delete return_type;
		delete parameters;
		delete definition;
	}

	virtual void build_entities ( CsDisplay & display );
	static void eval_and_check_entity_flags ( CsDisplay & display, CsModifiers & modifiers, CsEntity * entity );
	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void addEntityDefinitions ( CsEntityCollector & tokenCollector );
	virtual void addEntityReferences ( CsEntityCollector & tokenCollector );
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
		if ( attributes )
			attributes->visit ( visitor );
		if ( return_type )
			return_type->visit ( visitor );
		if ( parameters )
			parameters->visit ( visitor );
		if ( definition )
			definition->visit ( visitor );
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	CsTypeRef * return_type;
	CsCompoundIdentifier compound_identifier;
	CsFormalParameterList * parameters;
	CsBlock * definition;
	CsEntityMethod * entity;
	CsTokenIndex lparen_token;
	CsTokenIndex rparen_token;
	CsTokenIndex semicolon_token;
};

DEFINE_AST_ALLOCATOR(CsMethod)

//------------------------------------------------------------------------------
// CsPropertyAccessor
//------------------------------------------------------------------------------

struct CsPropertyAccessor : public CsNode
{
	CsPropertyAccessor()
		: CsNode ( n_property_accessor )
		, attributes(0)
		, definition(0)
		, entity(0)
		, get_set_token(0)
		, semicolon_token(0)
	{}

	explicit CsPropertyAccessor ( LineInfo line_info )
		: CsNode ( n_property_accessor, line_info )
		, attributes(0)
		, definition(0)
		, entity(0)
		, get_set_token(0)
		, semicolon_token(0)
	{}

	virtual ~CsPropertyAccessor()
	{
		delete attributes;
		delete definition;
	}

	virtual void build_entities ( CsDisplay & display );
	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void addEntityDefinitions ( CsEntityCollector & tokenCollector );
	virtual void addEntityReferences ( CsEntityCollector & tokenCollector );
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
		if ( attributes )
			attributes->visit ( visitor );
		if ( definition )
			definition->visit ( visitor );
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	CsAttributes * attributes;
	bool bSet; // true - set; false - get
	CsBlock * definition;
	CsEntityMethod * entity;
	CsTokenIndex get_set_token;
	CsTokenIndex semicolon_token;
};

DEFINE_AST_ALLOCATOR(CsPropertyAccessor)

//------------------------------------------------------------------------------
// CsProperty
//------------------------------------------------------------------------------

struct CsProperty : public CsNode, public CsAttributesModifiers
{
	CsProperty()
		: CsNode ( n_property_declaration )
		, get(0)
		, set(0)
		, entity(0)
		, lcurly_token(0)
		, rcurly_token(0)
		, semicolon_token(0)
	{}

	explicit CsProperty ( LineInfo line_info )
		: CsNode ( n_property_declaration, line_info )
		, get(0)
		, set(0)
		, entity(0)
		, lcurly_token(0)
		, rcurly_token(0)
		, semicolon_token(0)
	{}

	virtual ~CsProperty()
	{
		delete type;
		delete get;
		delete set;
	}

	virtual void build_entities ( CsDisplay & display );
	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void addEntityDefinitions ( CsEntityCollector & tokenCollector );
	virtual void addEntityReferences ( CsEntityCollector & tokenCollector );
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
		if ( attributes )
			attributes->visit ( visitor );
		if ( type )
			type->visit ( visitor );
		if ( get )
			get->visit ( visitor );
		if ( set )
			set->visit ( visitor );
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	CsTypeRef * type;
	CsCompoundIdentifier compound_identifier;
	CsPropertyAccessor * get;
	CsPropertyAccessor * set;
	CsEntityProperty * entity;
	CsTokenIndex lcurly_token;
	CsTokenIndex rcurly_token;
	CsTokenIndex semicolon_token;
};

DEFINE_AST_ALLOCATOR(CsProperty)

//------------------------------------------------------------------------------
// CsEventAccessor
//------------------------------------------------------------------------------

struct CsEventAccessor : public CsNode, public CsAttributesModifiers
{
	CsEventAccessor()
		: CsNode ( n_event_accessor )
		, definition(0)
		, entity(0)
		, add_remove_token(0)
		, semicolon_token(0)
	{}

	explicit CsEventAccessor ( LineInfo line_info )
		: CsNode ( n_event_accessor, line_info )
		, definition(0)
		, entity(0)
		, add_remove_token(0)
		, semicolon_token(0)
	{}

	virtual ~CsEventAccessor()
	{
		delete definition;
	}

	virtual void build_entities ( CsDisplay & display );
	static void type_lookup ( CsDisplay & display, CsEntityMethod * method, CsTypeRef * type );
	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void addEntityDefinitions ( CsEntityCollector & tokenCollector );
	virtual void addEntityReferences ( CsEntityCollector & tokenCollector );
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
		if ( attributes )
			attributes->visit ( visitor );
		if ( definition )
			definition->visit ( visitor );
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	bool bAdd; // true = add, false - remove
	CsBlock * definition;
	CsEntityMethod * entity;
	CsTokenIndex add_remove_token;
	CsTokenIndex semicolon_token;
};

DEFINE_AST_ALLOCATOR(CsEventAccessor)

//------------------------------------------------------------------------------
// CsEventDeclarator
//------------------------------------------------------------------------------

struct CsEventDeclarator : public CsNode
{
	CsEventDeclarator()
		: CsNode ( n_event_declarator )
		, expression(0)
		, add(0)
		, remove(0)
		, entity(0)
		, separator_token(0)
		, assign_token(0)
		, lcurly_token(0)
		, rcurly_token(0)
	{}

	explicit CsEventDeclarator ( LineInfo line_info )
		: CsNode ( n_event_declarator, line_info )
		, expression(0)
		, add(0)
		, remove(0)
		, entity(0)
		, separator_token(0)
		, assign_token(0)
		, lcurly_token(0)
		, rcurly_token(0)
	{}

	
	CsEventDeclarator ( LineInfo line_info, WPHashString phs, CsTokenIndex identifier_token )
		: CsNode ( n_event_declarator, line_info )
		, compound_identifier ( phs, identifier_token )
		, expression(0)
		, add(0)
		, remove(0)
		, entity(0)
		, separator_token(0)
		, assign_token(0)
		, lcurly_token(0)
		, rcurly_token(0)
	{}


	virtual ~CsEventDeclarator()
	{
		delete expression;
		delete add;
		delete remove;
	}

	virtual void build_entities ( CsDisplay & display );
	CsEntityMethod * generate_event_accessor ( const char * prefix, CsDisplay & display );
	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void addEntityDefinitions ( CsEntityCollector & tokenCollector );
	virtual void addEntityReferences ( CsEntityCollector & tokenCollector );
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
		if ( add )
			add->visit ( visitor );
		if ( remove )
			remove->visit ( visitor );
		if ( expression )
			expression->visit ( visitor );
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	CsCompoundIdentifier compound_identifier;
	CsEventAccessor * add;
	CsEventAccessor * remove;
	CsExpression * expression;
	CsEntityEvent * entity;
	CsTokenIndex separator_token;
	CsTokenIndex assign_token;
	CsTokenIndex lcurly_token;
	CsTokenIndex rcurly_token;
};

DEFINE_AST_ALLOCATOR(CsEventDeclarator)

//------------------------------------------------------------------------------
// CsEvent
//------------------------------------------------------------------------------

struct CsEvent : public CsNode, public CsAttributesModifiers
{
	CsEvent()
		: CsNode ( n_event_declaration, line_info )
		, type(0)
		, event_token(0)
		, semicolon_token(0)
	{}

	explicit CsEvent ( LineInfo line_info )
		: CsNode ( n_event_declaration, line_info )
		, type(0)
		, event_token(0)
		, semicolon_token(0)
	{}

	CsEvent ( cs_node type, LineInfo line_info )
		: CsNode ( type, line_info )
		, type(0)
		, event_token(0)
		, semicolon_token(0)
	{}

	virtual ~CsEvent()
	{
		delete type;
		declarators.delete_nodes_and_values();
	}

	virtual void build_entities ( CsDisplay & display );
	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void addEntityDefinitions ( CsEntityCollector & tokenCollector ) {}
	virtual void addEntityReferences ( CsEntityCollector & tokenCollector ) {}
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
		if ( attributes )
			attributes->visit ( visitor );
		if ( type )
			type->visit ( visitor );
		const single_list1<CsEventDeclarator*>::node * next
			= & declarators.first_;
		while ( next )
		{
			if ( next->value_ )
				next->value_->visit ( visitor );
			next = next->next_;
		}
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	CsTypeRef * type;
	single_list1<CsEventDeclarator*> declarators;
	CsTokenIndex event_token;
	CsTokenIndex semicolon_token;
};

DEFINE_AST_ALLOCATOR(CsEvent)

//------------------------------------------------------------------------------
// CsIndexer
//------------------------------------------------------------------------------

struct CsIndexer : public CsNode, public CsAttributesModifiers
{
	CsIndexer()
		: CsNode ( n_indexer_declaration )
		, type(0)
		, parameters(0)
		, get(0)
		, set(0)
		, entity(0)
		, dot_token(0)
		, this_token(0)
		, lbracket_token(0)
		, rbracket_token(0)
		, lcurly_token(0)
		, rcurly_token(0)
		, semicolon_token(0)
	{}

	explicit CsIndexer ( LineInfo line_info )
		: CsNode ( n_indexer_declaration, line_info )
		, type(0)
		, parameters(0)
		, get(0)
		, set(0)
		, entity(0)
		, dot_token(0)
		, this_token(0)
		, lbracket_token(0)
		, rbracket_token(0)
		, lcurly_token(0)
		, rcurly_token(0)
		, semicolon_token(0)
	{}

	virtual ~CsIndexer()
	{

		delete type;
		delete parameters;
		delete get;
		delete set;
	}

	virtual void build_entities ( CsDisplay & display );
	PHashString get_last_name ( CsDisplay & display );
	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void addEntityDefinitions ( CsEntityCollector & tokenCollector );
	virtual void addEntityReferences ( CsEntityCollector & tokenCollector );
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
		if ( attributes )
			attributes->visit ( visitor );
		if ( type )
			type->visit ( visitor );
		if ( parameters )
			parameters->visit ( visitor );
		if ( get )
			get->visit ( visitor );
		if ( set )
			set->visit ( visitor );
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	CsTypeRef * type;
	CsCompoundIdentifier compound_prefix;
	CsFormalParameterList * parameters;
	CsPropertyAccessor * get;
	CsPropertyAccessor * set;
	CsEntityProperty * entity;
	CsTokenIndex dot_token;
	CsTokenIndex this_token;
	CsTokenIndex lbracket_token;
	CsTokenIndex rbracket_token;
	CsTokenIndex lcurly_token;
	CsTokenIndex rcurly_token;
	CsTokenIndex semicolon_token;
};

DEFINE_AST_ALLOCATOR(CsIndexer)

//------------------------------------------------------------------------------
// CsOperator
//------------------------------------------------------------------------------

struct CsOperator : public CsNode, public CsAttributesModifiers
{
	CsOperator()
		: CsNode ( n_operator_declaration )
		, type(0)
		, parameter1(0)
		, parameter2(0)
		, definition(0)
		, entity(0)
		, operator_token(0)
		, operation_token(0)
		, lparen_token(0)
		, comma_token(0)
		, rparen_token(0)
		, semicolon_token(0)
	{}

	explicit CsOperator ( LineInfo line_info )
		: CsNode ( n_operator_declaration, line_info )
		, type(0)
		, parameter1(0)
		, parameter2(0)
		, definition(0)
		, entity(0)
		, operator_token(0)
		, operation_token(0)
		, lparen_token(0)
		, comma_token(0)
		, rparen_token(0)
		, semicolon_token(0)
	{}

	virtual ~CsOperator()
	{
		delete type;
		delete parameter1;
		delete parameter2;
		delete definition;
	}

	virtual void build_entities ( CsDisplay & display );
	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void addEntityDefinitions ( CsEntityCollector & tokenCollector );
	virtual void addEntityReferences ( CsEntityCollector & tokenCollector );
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
		if ( attributes )
			attributes->visit ( visitor );
		if ( type )
			type->visit ( visitor );
		if ( parameter1 )
			parameter1->visit ( visitor );
		if ( parameter2 )
			parameter2->visit ( visitor );
		if ( definition )
			definition->visit ( visitor );
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	CsTypeRef * type;
	CsTokenType operation;
	CsFormalParameter * parameter1;
	CsFormalParameter * parameter2;
	CsBlock * definition;
	CsEntityMethod * entity;
	CsTokenIndex operator_token;
	CsTokenIndex operation_token;
	CsTokenIndex lparen_token;
	CsTokenIndex comma_token;
	CsTokenIndex rparen_token;
	CsTokenIndex semicolon_token;
};

DEFINE_AST_ALLOCATOR(CsOperator)

//------------------------------------------------------------------------------
// CsConversionOperatorDeclaration
//------------------------------------------------------------------------------

struct CsConversionOperatorDeclaration : public CsNode, public CsAttributesModifiers
{
	CsConversionOperatorDeclaration()
		: CsNode ( n_conversion_operator_declaration )
		, type(0)
		, parameter1(0)
		, definition(0)
		, entity(0)
		, explicit_implicit_token(0)
		, operator_token(0)
		, lparen_token(0)
		, rparen_token(0)
		, semicolon_token(0)
	{}

	explicit CsConversionOperatorDeclaration ( LineInfo line_info )
		: CsNode ( n_conversion_operator_declaration, line_info )
		, type(0)
		, parameter1(0)
		, definition(0)
		, entity(0)
		, explicit_implicit_token(0)
		, operator_token(0)
		, lparen_token(0)
		, rparen_token(0)
		, semicolon_token(0)
	{}

	virtual ~CsConversionOperatorDeclaration()
	{
		delete type;
		delete parameter1;
		delete definition;
	}

	virtual void build_entities ( CsDisplay & display );
	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void addEntityDefinitions ( CsEntityCollector & tokenCollector );
	virtual void addEntityReferences ( CsEntityCollector & tokenCollector );
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
		if ( attributes )
			attributes->visit ( visitor );
		if ( type )
			type->visit ( visitor );
		if ( parameter1 )
			parameter1->visit ( visitor );
		if ( definition )
			definition->visit ( visitor );
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	CsTokenType modifier; // explicit/implicit
	CsTypeRef * type;
	CsFormalParameter * parameter1;
	CsBlock * definition;
	CsEntityMethod * entity;
	CsTokenIndex explicit_implicit_token;
	CsTokenIndex operator_token;
	CsTokenIndex lparen_token;
	CsTokenIndex rparen_token;
	CsTokenIndex semicolon_token;
};

DEFINE_AST_ALLOCATOR(CsConversionOperatorDeclaration)

//------------------------------------------------------------------------------
// CsDestructorDeclaration
//------------------------------------------------------------------------------

struct CsDestructorDeclaration : public CsNode, public CsAttributesModifiers
{
	CsDestructorDeclaration()
		: CsNode ( n_destructor_declaration )
		, definition(0)
		, entity(0)
		, tilde_token(0)
		, identifier_token(0)
		, lparen_token(0)
		, rparen_token(0)
		, semicolon_token(0)
	{}

	explicit CsDestructorDeclaration ( LineInfo line_info )
		: CsNode ( n_destructor_declaration, line_info )
		, definition(0)
		, entity(0)
		, tilde_token(0)
		, identifier_token(0)
		, lparen_token(0)
		, rparen_token(0)
		, semicolon_token(0)
	{}

	virtual ~CsDestructorDeclaration()
	{
		delete definition;
	}

	virtual void build_entities ( CsDisplay & display );
	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void addEntityDefinitions ( CsEntityCollector & tokenCollector );
	virtual void addEntityReferences ( CsEntityCollector & tokenCollector );
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
		if ( attributes )
			attributes->visit ( visitor );
		if ( definition )
			definition->visit ( visitor );
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	PHashString name;
	CsBlock * definition;
	CsEntityMethod * entity;
	CsTokenIndex tilde_token;
	CsTokenIndex identifier_token;
	CsTokenIndex lparen_token;
	CsTokenIndex rparen_token;
	CsTokenIndex semicolon_token;
};

DEFINE_AST_ALLOCATOR(CsDestructorDeclaration)

//------------------------------------------------------------------------------
// CsCompilationUnit
//------------------------------------------------------------------------------

struct CsCompilationUnit : public CsScopeNodeWithUsingDirectives
{
	CsCompilationUnit()
		: CsScopeNodeWithUsingDirectives ( n_compilation_unit )
		, file ( 0 )
	{}

	virtual ~CsCompilationUnit()
	{
		DELETE_FROM_CONTAINER(FastArray<CsAttributes*>,attributes);
		DELETE_FROM_CONTAINER(FastArray<CsNode*>,declarations);
	}

	virtual void build_entities ( CsDisplay & display );
	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void addEntityDefinitions ( CsEntityCollector & tokenCollector ) {}
	virtual void addEntityReferences ( CsEntityCollector & tokenCollector ) {}
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
		CsScopeNodeWithUsingDirectives::visit ( visitor );
		{
			FastArray<CsAttributes*>::iterator it = attributes.begin();
			FastArray<CsAttributes*>::iterator it_end = attributes.end();
			for ( ; it != it_end; ++it )
			{
				if ( *it )
					(*it)->visit ( visitor );
			}
		}
		{
			FastArray<CsNode*>::iterator it = declarations.begin();
			FastArray<CsNode*>::iterator it_end = declarations.end();
			for ( ; it != it_end; ++it )
			{
				if ( *it )
					(*it)->visit ( visitor );
			}
		}
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	FastArray<CsAttributes*> attributes;
	FastArray<CsNode*> declarations;
	CsFile * file;
};

DEFINE_AST_ALLOCATOR(CsCompilationUnit)

//------------------------------------------------------------------------------
// CsMakeRefExpression
//------------------------------------------------------------------------------

struct CsMakeRefExpression : public CsPrimaryExpression
{
	CsMakeRefExpression()
		: CsPrimaryExpression ( n_makeref )
		, expression(0)
		, makeref_token(0)
		, lparen_token(0)
		, rparen_token(0)
	{}

	explicit CsMakeRefExpression ( LineInfo line_info )
		: CsPrimaryExpression  ( n_makeref, line_info )
		, expression(0)
		, makeref_token(0)
		, lparen_token(0)
		, rparen_token(0)
	{}

	virtual ~CsMakeRefExpression()
	{
		delete expression;
	}

	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void addEntityDefinitions ( CsEntityCollector & tokenCollector ) {}
	virtual void addEntityReferences ( CsEntityCollector & tokenCollector ) {}
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
		if ( expression )
			expression->visit ( visitor );
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	CsExpression * expression;
	CsTokenIndex makeref_token;
	CsTokenIndex lparen_token;
	CsTokenIndex rparen_token;
};

DEFINE_AST_ALLOCATOR(CsMakeRefExpression)

//------------------------------------------------------------------------------
// CsRefTypeExpression
//------------------------------------------------------------------------------

struct CsRefTypeExpression : public CsPrimaryExpression
{
	CsRefTypeExpression()
		: CsPrimaryExpression ( n_reftype )
		, expression(0)
		, reftype_token(0)
		, lparen_token(0)
		, rparen_token(0)
	{}

	explicit CsRefTypeExpression ( LineInfo line_info )
		: CsPrimaryExpression ( n_reftype, line_info )
		, expression(0)
		, reftype_token(0)
		, lparen_token(0)
		, rparen_token(0)
	{}

	virtual ~CsRefTypeExpression()
	{
		delete expression;
	}

	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void addEntityDefinitions ( CsEntityCollector & tokenCollector ) {}
	virtual void addEntityReferences ( CsEntityCollector & tokenCollector ) {}
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
		if ( expression )
			expression->visit ( visitor );
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	CsExpression * expression;
	CsTokenIndex reftype_token;
	CsTokenIndex lparen_token;
	CsTokenIndex rparen_token;
};

DEFINE_AST_ALLOCATOR(CsRefTypeExpression)

//------------------------------------------------------------------------------
// CsRefValueExpression
//------------------------------------------------------------------------------

struct CsRefValueExpression : public CsPrimaryExpression
{
	CsRefValueExpression()
		: CsPrimaryExpression ( n_refvalue )
		, expression(0)
		, type(0)
		, refvalue_token(0)
		, lparen_token(0)
		, comma_token(0)
		, rparen_token(0)
	{}

	explicit CsRefValueExpression ( LineInfo line_info )
		: CsPrimaryExpression ( n_refvalue, line_info )
		, expression(0)
		, type(0)
		, refvalue_token(0)
		, lparen_token(0)
		, comma_token(0)
		, rparen_token(0)
	{}

	virtual ~CsRefValueExpression()
	{
		delete expression;
		delete type;
	}

	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void addEntityDefinitions ( CsEntityCollector & tokenCollector ) {}
	virtual void addEntityReferences ( CsEntityCollector & tokenCollector ) {}
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
		if ( expression )
			expression->visit ( visitor );
		if ( type )
			type->visit ( visitor );
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	CsExpression * expression;
	CsTypeRef * type;
	CsTokenIndex refvalue_token;
	CsTokenIndex lparen_token;
	CsTokenIndex comma_token;
	CsTokenIndex rparen_token;
};

DEFINE_AST_ALLOCATOR(CsRefValueExpression)

//------------------------------------------------------------------------------
// CsArgListExpression
//------------------------------------------------------------------------------

struct CsArgListExpression : public CsPrimaryExpression
{
	CsArgListExpression()
		: CsPrimaryExpression ( n_arglist )
		, expression_list(0)
		, arglist_token(0)
		, lparen_token(0)
		, rparen_token(0)
	{}

	explicit CsArgListExpression ( LineInfo line_info )
		: CsPrimaryExpression ( n_arglist, line_info )
		, expression_list(0)
		, arglist_token(0)
		, lparen_token(0)
		, rparen_token(0)
	{}

	virtual ~CsArgListExpression()
	{
		delete expression_list;
	}

	virtual void type_lookup ( CsDisplay & display );
	virtual void member_lookup ( CsDisplay & display );

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector );
	virtual void addTokens ( CsTokenCollector & tokenCollector );
	virtual void addEntityDefinitions ( CsEntityCollector & tokenCollector ) {}
	virtual void addEntityReferences ( CsEntityCollector & tokenCollector ) {}
	virtual void clearEntities();


	virtual void visit ( CsAstVisitor * visitor )
	{
		visitor->visit ( this );
		if ( expression_list )
			expression_list->visit ( visitor );
	}

	virtual void callback ( CsAstCallback * visitor )
	{
		visitor->callback ( this );
	}

	static inline void * operator new ( size_t size );
	static inline void operator delete ( void * address );

	CsExpressionList * expression_list;
	CsTokenIndex arglist_token;
	CsTokenIndex lparen_token;
	CsTokenIndex rparen_token;
};

DEFINE_AST_ALLOCATOR(CsArgListExpression)

//------------------------------------------------------------------------------
// CsStuffExpression
//------------------------------------------------------------------------------

struct CsStuffExpression : public CsExpression
{
	CsStuffExpression()
		: CsExpression ( n_null, LineInfo() )
	{}


	virtual void visit ( CsAstVisitor * visitor )
	{
	}

	virtual void callback ( CsAstCallback * visitor )
	{
	}

	virtual void addTokensRecursive ( CsTokenCollector & tokenCollector )
	{
	}
	virtual void addTokens ( CsTokenCollector & tokenCollector )
	{
	}

	virtual void type_lookup ( CsDisplay & display )
	{
	}
	virtual void member_lookup ( CsDisplay & display )
	{
	}
};


#endif // _CS_PARSER_TYPES_H_
